
#include "EquivalenceConstructor.h"

#include <stdexcept>
#include <stdio.h>
#include <string.h>
//#include "libxml/encoding.h"
//#include "libxml/xmlwriter.h"
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>


bool operator< (const NodeInfo& n1, const NodeInfo& n2)
{
	return (n1.id<n2.id);
}


xmlNodePtr EquivalenceConstructor:: GetChildrenNode(xmlNodePtr parent, char* name_child)
{
	xmlNodePtr child = parent->xmlChildrenNode;
	while((child != NULL) && (xmlStrcasecmp(child->name, (const xmlChar*)name_child)))
	{
		child = child->next;
	}
	
	return child;
	
}

xmlNodePtr EquivalenceConstructor:: GetNextChild(xmlNodePtr child1, char* name_child)
{
	
	xmlNodePtr child2 = child1->next;
	while((child2 != NULL) && (xmlStrcasecmp(child2->name, (const xmlChar*)name_child)))
	{
		child2 = child2->next;
	}
	
	return child2;
	
	
}

void EquivalenceConstructor:: ReadHypergraphInfo(xmlNodePtr HyperInfo, HypergraphInfo &hgInfo)
{
try
{
	string str;
	
	str = (char *)xmlGetProp(HyperInfo, (const xmlChar*) "id");
	hgInfo.id = atoi(str.c_str());

	str = (char *)xmlGetProp(HyperInfo, (const xmlChar*) "link");
	hgInfo.link = atoi(str.c_str());
	
	xmlNodePtr HypGraph = NULL;
	HypGraph = GetChildrenNode(HyperInfo, (char*)"Hypergraph");
	if (HypGraph != NULL)
	{
		ReadHypergraph(HypGraph, hgInfo.hg);
	}
	
	xmlNodePtr MatchesLG_XmlPointer = NULL;
	MatchesLG_XmlPointer = GetChildrenNode(HyperInfo, (char*)"matchesLG");
	
	
	//if (xmlStrcasecmp(MatchesLG_XmlPointer->name, (const xmlChar*) "matchesLG"))
	if(MatchesLG_XmlPointer == NULL)
      		throw runtime_error("matchesLG expected");
		
	typedef pair<IDTYPE, IDTYPE> matches_type;
	map<IDTYPE, IDTYPE> map_el;
	map_el.clear();
	IDTYPE id1, id2;
	
	xmlNodePtr list_element = NULL;
	list_element = GetChildrenNode(MatchesLG_XmlPointer, (char*)"map");
	while(list_element != NULL)
	{
		xmlNodePtr map_element = NULL;
		map_element = GetChildrenNode(list_element,(char*)"map_element");
		while(map_element !=NULL)
		{
			str = (char *)xmlGetProp(map_element, (const xmlChar*) "id1");
			id1 = atoi(str.c_str());
			str = (char *)xmlGetProp(map_element, (const xmlChar*) "id2");
			id2 = atoi(str.c_str());
			map_el.insert(matches_type(id1,id2));
			map_element = GetNextChild(map_element, (char*)"map_element");
		}
		
		hgInfo.matchesLG.push_back(map_el);
		map_el.clear();
		list_element = GetNextChild(list_element,(char*)"map");
	}
	
	
	xmlNodePtr MatchesGgM_XmlPointer = NULL;
	MatchesGgM_XmlPointer = GetChildrenNode(HyperInfo, (char*)"matchesGgM");
	
	//if (xmlStrcasecmp(cur->name, (const xmlChar*) "matchesGgM"))
	if(MatchesGgM_XmlPointer == NULL)
      		throw runtime_error("matchesGgM expected");
	
	list_element = NULL;	
	list_element = GetChildrenNode(MatchesGgM_XmlPointer,(char*)"map");
	map_el.clear();
	while(list_element != NULL)
	{
		xmlNodePtr map_element = NULL;
		map_element = GetChildrenNode(list_element,(char*)"map_element");
		while(map_element != NULL)
		{
			str = (char *)xmlGetProp(map_element, (const xmlChar*) "id1");
			id1 = atoi(str.c_str());
			str = (char *)xmlGetProp(map_element, (const xmlChar*) "id2");
			id2 = atoi(str.c_str());
			map_el.insert(matches_type(id1,id2));
			map_element = GetNextChild(map_element, (char*)"map_element");
		}
		hgInfo.matchesGgM.push_back(map_el);
		map_el.clear();
		list_element = GetNextChild(list_element, (char*)"map");
	}
}
catch(exception e)
{ throw e;}
}

// write HypergraphInfo into xml-file

void EquivalenceConstructor:: WriteHypergraphInfo(xmlNodePtr list_el, HypergraphInfo &hgInfo)
{
	char ch[25];
	xmlNodePtr HyperInfo = xmlNewChild(list_el, NULL, (const xmlChar *)"HypergraphInfo", NULL);
	sprintf(ch,"%i",hgInfo.id);
	xmlSetProp(HyperInfo,(const xmlChar*)"id",(const xmlChar*)ch);
	sprintf(ch,"%i",hgInfo.link);
	xmlSetProp(HyperInfo,(const xmlChar*)"link",(const xmlChar*)ch);

	WriteHypergraph(HyperInfo, hgInfo.hg);

	xmlNodePtr matchesLG = xmlNewChild(HyperInfo, NULL,(const xmlChar*)"matchesLG", NULL);
	for(list<map<IDTYPE,IDTYPE> >::iterator it = hgInfo.matchesLG.begin(); it!= hgInfo.matchesLG.end(); it++)
	{
		xmlNodePtr map_matchesLG = xmlNewChild(matchesLG, NULL,(const xmlChar*)"map",NULL);	
		for(map<IDTYPE,IDTYPE>::iterator itm = it->begin(); itm!= it->end(); itm++)	
		{
			xmlNodePtr map_element = xmlNewChild(map_matchesLG,NULL,(const xmlChar*)"map_element",NULL);
			sprintf(ch,"%i",itm->first);
			xmlSetProp(map_element,(const xmlChar*)"id1",(const xmlChar*)ch);
			sprintf(ch,"%i",itm->second);
			xmlSetProp(map_element,(const xmlChar*)"id2",(const xmlChar*)ch);
		}
	}
	xmlNodePtr matchesGgM = xmlNewChild(HyperInfo, NULL,(const xmlChar*)"matchesGgM", NULL);
	for(list<map<IDTYPE,IDTYPE> >::iterator it = hgInfo.matchesGgM.begin(); it!= hgInfo.matchesGgM.end(); it++)
	{
		xmlNodePtr map_matchesGgM = xmlNewChild(matchesGgM, NULL,(const xmlChar*)"map",NULL);	
		for(map<IDTYPE,IDTYPE>::iterator itm = it->begin(); itm!= it->end(); itm++)	
		{
			xmlNodePtr map_element = xmlNewChild(map_matchesGgM,NULL,(const xmlChar*)"map_element",NULL);
			sprintf(ch,"%i",itm->first);
			xmlSetProp(map_element,(const xmlChar*)"id1",(const xmlChar*)ch);
			sprintf(ch,"%i",itm->second);
			xmlSetProp(map_element,(const xmlChar*)"id2",(const xmlChar*)ch);
		}
	}
}

void EquivalenceConstructor:: ReadHypergraph(xmlNodePtr HypGraph, Hypergraph &hg)
{
  char *ch;
  string str;
try
{
  // set pointer on vertices
  xmlNodePtr hg_element = NULL;
  hg_element = GetChildrenNode(HypGraph, (char*)"vertices");

  //if (xmlStrcasecmp(hg_element->name, (const xmlChar*) "vertices"))
  if(hg_element == NULL)
      throw runtime_error("vertices expected");
 
  xmlNodePtr cur = NULL;
  cur = hg_element->xmlChildrenNode;
  if (cur != NULL)
  {
	str = (char*)cur->content; 
	if (str != "")
    	{
			ch = strtok((char*)str.c_str(), ";");

			while(ch!=NULL)
			{
				hg.vertices.insert(atoi(ch));
				ch = strtok(NULL, ";");
			}
	}
  }
  
// set pointer on edges	
  hg_element = GetChildrenNode(HypGraph,(char*)"edges");

  //if (xmlStrcasecmp(hg_element->name, (const xmlChar*) "edges"))
  if(hg_element == NULL)
      throw runtime_error("edges expected");
      
  cur = NULL;
  cur = hg_element->xmlChildrenNode;
  if(cur != NULL)
  {
  	str = (char*)cur->content; 
      	if (str != "")
    	{
			ch = strtok((char*)str.c_str(), ";");

			while(ch!=NULL)
			{
				hg.edges.insert(atoi(ch));
				ch = strtok(NULL, ";");
			}
	}
  }
  
// set pointer on connectedV	
  hg_element = GetChildrenNode(HypGraph,(char*)"connectedV");

  //if (xmlStrcasecmp(hg_element->name, (const xmlChar*) "connectedV"))
  if(hg_element == NULL)
      throw runtime_error("ConnectedV expected");
      
  xmlNodePtr map_element = NULL;
  map_element = GetChildrenNode(hg_element,(char*)"map_element");
  
  typedef pair<IDTYPE, list<IDTYPE> > connectedV_type;
  list<IDTYPE> lst;
  IDTYPE id;
  
  lst.clear();
  while (map_element != NULL)
  {
      str = (char *)xmlGetProp(map_element, (const xmlChar*) "id");
      id = atoi(str.c_str());
      cur = NULL;
      cur = map_element->xmlChildrenNode;
      
      if(cur != NULL) 
      {
      	str = (char*)cur->content;
 
      	if (str != "")
    	{
			ch = strtok((char*)str.c_str(), ";");
			while(ch!=NULL)
			{	
				lst.push_back(atoi(ch));
				ch = strtok(NULL, ";");
			}
			
	}
      }
	
      hg.connectedV.insert( connectedV_type(id, lst) );
      
      lst.clear();
    
      map_element = GetNextChild(map_element,(char*)"map_element");
     
    }
     
    
   // set pointer on connectedE	
  hg_element = GetChildrenNode(HypGraph,(char*)"connectedE");

 // if (xmlStrcasecmp(hg_element->name, (const xmlChar*) "connectedE"))
 if(hg_element == NULL)
      throw runtime_error("ConnectedE expected");
      
  map_element = GetChildrenNode(hg_element,(char*)"map_element");
  
  typedef pair<IDTYPE, multiset<IDTYPE> > connectedE_type;
  multiset<IDTYPE> set_E;
  
  set_E.clear();
  
  while (map_element != NULL)
  {
      str = (char *)xmlGetProp(map_element, (const xmlChar*) "id");
      id = atoi(str.c_str());
      cur = map_element->xmlChildrenNode;
      if(cur != NULL)
      {
      	str = (char*)cur->content;
		
      	if (str != "")
    	{
			ch = strtok((char*)str.c_str(), ";");

			while(ch!=NULL)
			{
				set_E.insert(atoi(ch));
				ch = strtok(NULL, ";");
				
			}
	}
      }
      
      hg.connectedE.insert(connectedE_type(id, set_E) );
      set_E.clear();
      
      map_element = GetNextChild(map_element, (char*)"map_element");
    }
    
    
  // set pointer on labels  
  hg_element = GetChildrenNode(HypGraph,(char*)"labels");

  if(hg_element == NULL)
      throw runtime_error("labels expected");
      
  map_element = GetChildrenNode(hg_element,(char*)"map_element");
  
  typedef pair<IDTYPE, LABELTYPE> labels_element;
  
  while(map_element !=NULL)
  {
  	str = (char *)xmlGetProp(map_element, (const xmlChar*) "id");
	id = atoi(str.c_str());
	str = (char *)xmlGetProp(map_element, (const xmlChar*) "type");
	hg.labels.insert(labels_element(id, str));
	map_element = GetNextChild(map_element,(char*)"map_element");
  }
  
  
  // set pointer on idLabels 	
  hg_element = GetChildrenNode(HypGraph,(char*)"idLabels");

  if(hg_element == NULL)
      throw runtime_error("idLabels expected");
      
  map_element = GetChildrenNode(hg_element,(char*) "map_element");
  
  while(map_element != NULL)
  {
  	str = (char *)xmlGetProp(map_element, (const xmlChar*) "id");
	id = atoi(str.c_str());
	str = (char *)xmlGetProp(map_element, (const xmlChar*) "type");
	hg.idLabels.insert(labels_element(id, str));
	map_element = GetNextChild(map_element,(char*)"map_element");
  }
  
  
  // set pointer on eqLabels 
  hg_element = GetChildrenNode(HypGraph,(char*)"eq_Labels");

  if(hg_element == NULL)
      throw runtime_error("eq_labels expected");
      
  map_element = GetChildrenNode(hg_element,(char*)"map_element");
     
  lst.clear();
  while (map_element != NULL)
  {
      str = (char *)xmlGetProp(map_element, (const xmlChar*) "id");
      id = atoi(str.c_str());
      
      xmlNodePtr cur_eq = GetChildrenNode(map_element,(char*)"list_element");
      
      while(cur_eq != NULL)
      {
      	cur = cur_eq->xmlChildrenNode;
	if (cur != NULL)
	{
      		ch = (char*)cur->content;
		lst.push_back(atoi(ch));
	}
	cur_eq = GetNextChild(cur_eq,(char*)"list_element");
      }
      
      hg.eq_labels.insert( connectedV_type(id, lst) );
      lst.clear();
      map_element = GetNextChild(map_element,(char*)"map_element");
  }
  
  
  // set pointer on depthmapEdges
  hg_element = GetChildrenNode(HypGraph,(char*)"depthmapEdges");

  if(hg_element == NULL)
      throw runtime_error("depthmapEdges expected");
      
  map_element = GetChildrenNode(hg_element,(char*)"map_element");
  
  typedef pair<IDTYPE, long> depthmap_type;
  IDTYPE id1;
  long id2;
  
  while(map_element != NULL)
  {
 	str = (char *)xmlGetProp(map_element, (const xmlChar*) "id");
	id1 = atoi(str.c_str());
	str = (char *)xmlGetProp(cur, (const xmlChar*) "type");
      	id2 = atol(str.c_str());
	hg.depthmapEdges.insert(depthmap_type(id1, id2));
	map_element = GetNextChild(map_element,(char*)"map_element");
  }
  
  
  // set pointer on depthmapVertices
  hg_element = GetChildrenNode(HypGraph,(char*)"depthmapVertices");

  if(hg_element == NULL)
      throw runtime_error("depthmapVertices expected");
      
  map_element = GetChildrenNode(hg_element,(char*)"map_element");
  
  while(map_element != NULL)
  {
 	str = (char *)xmlGetProp(map_element, (const xmlChar*) "id");
	id1 = atoi(str.c_str());
	str = (char *)xmlGetProp(cur, (const xmlChar*) "type");
      	id2 = atol(str.c_str());
	hg.depthmapVertices.insert(depthmap_type(id1, id2));
	map_element = GetNextChild(map_element,(char*)"map_element");
  }
  
  
// for empty
// for emptyIDTYPE
}
catch(exception e) 
{
	cout << "error by reading Hypergraph from XML file" << endl;
	throw e;
}
}               

void EquivalenceConstructor:: WriteHypergraph(xmlNodePtr HyperInfo, Hypergraph &hg)
{
  char ch[256];
  string s = "";
  xmlNodePtr Hgraph = xmlNewChild(HyperInfo, NULL,(const xmlChar*)"Hypergraph", NULL);
  
  set<IDTYPE> vertices_tmp = hg.vertices;

  for(set<IDTYPE>::iterator it = vertices_tmp.begin(); it!= vertices_tmp.end(); it++)
  {
      sprintf(ch,"%i",*it);
      s = s + ch +";";
  } 

 xmlNewChild(Hgraph, NULL, (const xmlChar*)"vertices", (const xmlChar*)s.c_str());
 
 s = "";
 set<IDTYPE> edges_tmp = hg.edges;
 for(set<IDTYPE>::iterator it = edges_tmp.begin(); it!= edges_tmp.end(); it++)
  {
      sprintf(ch,"%i",*it);
      s = s + ch +";";
  } 

 xmlNewChild(Hgraph, NULL, (const xmlChar*)"edges", (const xmlChar*)s.c_str());
 
  
  xmlNodePtr connectedV = xmlNewChild(Hgraph, NULL, (const xmlChar*)"connectedV", NULL);
  for(map<IDTYPE, list<IDTYPE> >::iterator it = hg.connectedV.begin(); it!= hg.connectedV.end(); it++)
  {
      s = "";	
      for(list<IDTYPE>::iterator itl = it->second.begin(); itl!= it->second.end(); itl++)
      {
      		sprintf(ch,"%i",*itl);
      		s = s + ch +";";
      }
      xmlNodePtr map_element = xmlNewChild(connectedV, NULL, (const xmlChar*)"map_element", (const xmlChar*)s.c_str());
      sprintf(ch, "%i",it->first);
      xmlSetProp(map_element,(const xmlChar*)"id",(const xmlChar*)ch);
      
   }
  
  xmlNodePtr connectedE = xmlNewChild(Hgraph, NULL, (const xmlChar*)"connectedE", NULL);
  for(map<IDTYPE, multiset<IDTYPE> >::iterator it = hg.connectedE.begin(); it!= hg.connectedE.end(); it++)
  {
  	s = "";	
      for(multiset<IDTYPE>::iterator itl = it->second.begin(); itl!= it->second.end(); itl++)
      {
      	sprintf(ch,"%i",*itl);
      	s = s + ch +";";
      
      }
      xmlNodePtr map_element = xmlNewChild(connectedE, NULL, (const xmlChar*)"map_element", (const xmlChar*)s.c_str());
      sprintf(ch, "%i",it->first);
      xmlSetProp(map_element,(const xmlChar*)"id",(const xmlChar*)ch);
   
  }

  const char *ch1;
  xmlNodePtr labels = xmlNewChild(Hgraph, NULL ,(const xmlChar*)"labels", NULL);
  for(map<IDTYPE,LABELTYPE>:: iterator it = hg.labels.begin(); it!= hg.labels.end(); it++)
  {
      xmlNodePtr map_element = xmlNewChild(labels, NULL, (const xmlChar*)"map_element", NULL);
      sprintf(ch,"%i", it->first);
      xmlSetProp(map_element, (const xmlChar*)"id", (const xmlChar*)ch);
      ch1 = it->second.c_str();
      xmlSetProp(map_element, (const xmlChar*)"type", (const xmlChar*)ch1);
  }

  xmlNodePtr idLabels = xmlNewChild(Hgraph, NULL ,(const xmlChar*)"idLabels", NULL);
  for(map<IDTYPE,std::string>:: iterator it = hg.idLabels.begin(); it!= hg.idLabels.end(); it++)
  {
      xmlNodePtr map_element = xmlNewChild(idLabels, NULL, (const xmlChar*)"map_element", NULL);
      sprintf(ch,"%i", it->first);
      xmlSetProp(map_element, (const xmlChar*)"id", (const xmlChar*)ch);
      ch1 = it->second.c_str();
      xmlSetProp(map_element, (const xmlChar*)"type", (const xmlChar*)ch1);
  }

  xmlNodePtr eq_labels = xmlNewChild(Hgraph, NULL ,(const xmlChar*)"eq_labels", NULL);
  for(map<IDTYPE,std::list<unsigned int> >:: iterator it = hg.eq_labels.begin(); it!= hg.eq_labels.end(); it++)
  {
      xmlNodePtr map_element = xmlNewChild(eq_labels, NULL, (const xmlChar*)"map_element", NULL);
      sprintf(ch,"%i", it->first);
      xmlSetProp(map_element, (const xmlChar*)"id", (const xmlChar*)ch);
      for(list<unsigned int>::iterator itl = it->second.begin(); itl!=it->second.end(); itl++)
      {
          sprintf(ch,"%i",*itl);
          xmlNewChild(map_element, NULL, (const xmlChar*)"list_element", (const xmlChar*)ch);
      }
  }

  xmlNodePtr depthmapEdges = xmlNewChild(Hgraph, NULL ,(const xmlChar*)"depthmapEdges", NULL);
  for(map<IDTYPE,long>:: iterator it = hg.depthmapEdges.begin(); it!= hg.depthmapEdges.end(); it++)
  {
      xmlNodePtr map_element = xmlNewChild(depthmapEdges, NULL, (const xmlChar*)"map_element", NULL);
      sprintf(ch,"%i", it->first);
      xmlSetProp(map_element, (const xmlChar*)"id", (const xmlChar*)ch);
      sprintf(ch,"%li", it->second);
      xmlSetProp(map_element, (const xmlChar*)"type", (const xmlChar*)ch);
  }

  xmlNodePtr depthmapVertices = xmlNewChild(Hgraph, NULL ,(const xmlChar*)"depthmapVertices", NULL);
  for(map<IDTYPE,long>:: iterator it = hg.depthmapVertices.begin(); it!= hg.depthmapVertices.end(); it++)
  {
      xmlNodePtr map_element = xmlNewChild(depthmapVertices, NULL, (const xmlChar*)"map_element", NULL);
      sprintf(ch,"%i", it->first);
      xmlSetProp(map_element, (const xmlChar*)"id", (const xmlChar*)ch);
      sprintf(ch,"%li", it->second);
      xmlSetProp(map_element, (const xmlChar*)"type", (const xmlChar*)ch);
  }

  xmlNodePtr empty = xmlNewChild(Hgraph, NULL, (const xmlChar*)"empty", NULL);
  for(list<unsigned int>::iterator it = hg.empty.begin(); it!= hg.empty.end(); it++)
  {
      sprintf(ch,"%i",*it);
      xmlNewChild(empty, NULL, (const xmlChar*)"list_element", (const xmlChar*)ch);
  }

  xmlNodePtr emptyIDTYPE = xmlNewChild(Hgraph, NULL, (const xmlChar*)"emptyIDTYPE", NULL);
  for(list<unsigned int>::iterator it = hg.emptyIDTYPE.begin(); it!= hg.emptyIDTYPE.end(); it++)
  {
      sprintf(ch,"%i",*it);
      xmlNewChild(emptyIDTYPE, NULL, (const xmlChar*)"list_element", (const xmlChar*)ch);
  }
}




void EquivalenceConstructor::Write_m_list(xmlNodePtr m_list_xml, vector<list<HypergraphInfo> > &m_list1)
{
	xmlNodePtr vector_m_list = xmlNewChild(m_list_xml, NULL,(const xmlChar *)"vector", NULL);

	for(vector<list<HypergraphInfo> >::iterator it = m_list1.begin(); it!=m_list1.end(); it++)
	{ 
		xmlNodePtr list_vector = xmlNewChild(vector_m_list, NULL,(const xmlChar *)"list", NULL);
		for(list<HypergraphInfo>::iterator itl = it->begin(); itl!= it->end(); itl++)
		{
			WriteHypergraphInfo(list_vector, *itl);	
		}	
	
	}

}




void EquivalenceConstructor::Read_m_list(xmlNodePtr m_list_xml, vector<list<HypergraphInfo> > &m_list1)
{ 
try
{
	xmlNodePtr list_el = NULL;	
	xmlNodePtr vector_el = NULL;
	vector_el = GetChildrenNode(m_list_xml, (char*)"vector");
	
	if (vector_el != NULL)
	{
		vector_el = GetChildrenNode(vector_el, (char*)"list");

		if (xmlStrcasecmp(vector_el->name, (const xmlChar*) "list"))
			throw runtime_error("Root node 'list' expected");

	
		list<HypergraphInfo> hgInfo_list;
		HypergraphInfo hygInfo;
	
		while(vector_el != NULL)
		{	
			list_el = GetChildrenNode(vector_el,(char*)"HypergraphInfo");
			while(list_el != NULL)
			{
				ReadHypergraphInfo(list_el, hygInfo);
				hgInfo_list.push_back(hygInfo);
				list_el = GetNextChild(list_el, (char*)"HypergraphInfo");
			}	
			m_list1.push_back(hgInfo_list);
			vector_el = GetNextChild(vector_el, (char*)"list");
		}
	
	}
}
catch(exception e)
{	
	cout << "error by reading HypergraphInfo from XML file " << endl; 
	throw e;
}
}


/** construct equvalence classes */
vector<list<HypergraphInfo> > EquivalenceConstructor::construct()
{
	  try
	  {
 	  // from empty level in m_list
	  const int level = m_list.size() - 2;

	  Hypergraph L = pg.getGraphGrammar().getRule(pg.getPTNet().getTransition(trans[level]).getAssociatedRule()).getLeft();

	  map<IDTYPE, IDTYPE> mu = pg.getMu(trans[level]);

	  set<IDTYPE> nodesL = L.getVertices();

		assert(m_list[level].size()!=0);

    for(int linklevel=level; linklevel>=0; linklevel--)
    {

	  // hypergraphs
	  for(list<HypergraphInfo>::iterator it = m_list[linklevel].begin(); it!=m_list[linklevel].end(); it++)
	  {
			// consider only if not linked
			if(isLinked(it->id, linklevel))
				continue;
			
      // assert that the matchings are initialized
      assert(it->matchesGgM.begin()!=it->matchesGgM.end());

				// matches G -> M
				for(list<map<IDTYPE, IDTYPE> >::iterator itmGM = it->matchesGgM.begin(); itmGM!=it->matchesGgM.end(); itmGM++)
				{
					// nodes
					for(set<IDTYPE>::iterator it_nodes=nodesL.begin(); it_nodes!=nodesL.end(); it_nodes++)
					{
						// class
						set<NodeInfo> eq_class;
						multiset<IDTYPE> edges = L.getConnectedEdges(*it_nodes);

						// edges
						for(set<IDTYPE>::iterator it_edges=edges.begin(); it_edges!=edges.end(); it_edges++)
						{
               IDTYPE id_edge_m = mu[*it_edges];

               // find pre-image in matching G -> M
               map<IDTYPE, IDTYPE>::iterator itr;
               
               for(itr = (*itmGM).begin(); itr!=(*itmGM).end(); itr++)
               {
								 if(itr->second == id_edge_m)
									 break;
               } 

               if(itr==(*itmGM).end())
                 throw runtime_error("internal error, matches");

               IDTYPE id_edge_g = itr->first;

               // order in L
							 list<IDTYPE> verticesL = L.getConnectedVertices(*it_edges);

							 // order in G
							 list<IDTYPE> verticesG = it->hg.getConnectedVertices(id_edge_g);

							 // find the corresponding node in G
							 list<IDTYPE>::iterator it_nG = verticesG.begin();
							 for(list<IDTYPE>::iterator it_nL = verticesL.begin(); it_nL!=verticesL.end(); it_nL++)
							 {
								 if(*it_nL == *it_nodes)
								 {
		 							 NodeInfo ni;
									 ni.id = *it_nG;
									 ni.graph_id = it->id;
									 ni.level = linklevel;

									 if(eq_class.find(ni)==eq_class.end())
										 eq_class.insert(ni);
								 } 

								 it_nG++;
							 }

						} // edges

            if(eq_class.size()>1)
            {
							// is the equivalence already in list
							bool is_in_list = false;
							for(list<set<NodeInfo> >::iterator itE = equivalences.begin(); itE!=equivalences.end(); itE++)
							{
                bool not_found = false;
								for(set<NodeInfo>::iterator itN = eq_class.begin(); itN!=eq_class.end(); itN++)
								{
									if(itE->find(*itN)==itE->end())
									{
										not_found = true;
										break;
									}	
								}

								if(!not_found)
								{
									is_in_list = true;
									break;
								}
							}

							if(!is_in_list)
								equivalences.push_back(eq_class);
						}	
	
					} // nodes

				} // GM

	  } // hypergraphs
	  
	  } // linklevel

	  // find appearing
	  findAppearing();

	  return m_list;
	  
	  }
	  catch(exception e)
	  {
		throw e;
	  }
}

/** find first appearing of each node in equivalence classes in [M_0] ... [M_n] */
void EquivalenceConstructor::findAppearing()
{
		try
		{
 	  // number of equivalence class
	  unsigned int eq_number = 1;

	  for(list<set<NodeInfo> >::iterator it_list = equivalences.begin();
						it_list!=equivalences.end(); it_list++)
		{
			for(set<NodeInfo>::iterator it = (*it_list).begin();
							it!=(*it_list).end(); it++)
			{
				int linklevel = it->level;
				HypergraphInfo * hgi = findHypergraph(it->graph_id, linklevel);
				HypergraphInfo * hgi_last;
				unsigned int graph_id = it->graph_id;
				unsigned int graph_id_last;
				
				while(linklevel>=0)
				{
					// G0
	  	  	if(linklevel==0)
					{
						hgi->hg.setEqLabel(it->id, eq_number);
						break; 
					}

          // copy last info
          hgi_last = hgi;
          graph_id_last = graph_id;
          
          graph_id = findLink(graph_id_last, linklevel);
          
          // find linked graph
          hgi = findHypergraph(graph_id, linklevel-1);

          // check if the node is still there
          set<IDTYPE> vertices = hgi->hg.getVertices();

          if(vertices.count(it->id)==0)
					{
						hgi_last->hg.setEqLabel(it->id, eq_number);
						break;
					}

          linklevel--;
				}
			}

			eq_number++;
		}

		optimize();
		
		}
		catch(exception e)
		{
			throw e;
		}
}

/** find in the level link of the graph with id */
unsigned int EquivalenceConstructor::findLink(unsigned int id, unsigned int level)
{
	  for(list<HypergraphInfo>::iterator it = m_list[level].begin(); it!=m_list[level].end(); it++)
	  {
			if(it->id==id)
				return it->link;
	  }

	  assert(false);
}

/** find hypergraph with id at the level */
HypergraphInfo * EquivalenceConstructor::findHypergraph(unsigned int id, unsigned int level)
{
 	  for(list<HypergraphInfo>::iterator it = m_list[level].begin(); it!=m_list[level].end(); it++)
    {
			if(it->id==id)
				return &(*it);
    }

	  assert(false);
}

/** optimize sample list by deleting the hypergraphs
		where no labels are marked
		and no labels will be marked at the next levels */
void EquivalenceConstructor::optimize()
{
	 for(int level = m_list.size() - 2; level >=0; level--)
	 {
		  list<int> to_delete;
		  
		  for(list<HypergraphInfo>::iterator it = m_list[level].begin(); it!=m_list[level].end(); it++)
		  {
				if(isLinked(it->id, level))
					continue;

				if(it->hg.getEqCount()==0)
					to_delete.push_back(it->id);
		  }

		  // to get no problems with deleted iterators let us make it so
		  for(list<int>::iterator itd = to_delete.begin(); itd!=to_delete.end(); itd++)
		  {
   		  for(list<HypergraphInfo>::iterator it = m_list[level].begin(); it!=m_list[level].end(); it++)
				{
					if(it->id == *itd)
					{
						m_list[level].erase(it);
						break;
					}
				}
		  }
				
	 }

 	 // delete empty sets
   int counter = 0;
   int count = m_list.size();
	 for(vector<list<HypergraphInfo> >::iterator it = m_list.begin(); it!=m_list.end(); it++)
	 {
		 if(it->size()==0)
		 {
       	 for(int i = 0; i<(count-counter); i++)
					 m_list.pop_back();

         break;
		 }

		 counter++;
	 }

	 // now propagate labels
   for(vector<list<HypergraphInfo> >::iterator itVec = m_list.begin(); itVec!=m_list.end(); itVec++)
	 {
		 vector<list<HypergraphInfo> >::iterator itTmp = itVec;

		 itTmp++;

  	 if(itTmp==m_list.end())
				break;
		 			 
		 for(list<HypergraphInfo>::iterator it1 = (*itVec).begin(); it1!=(*itVec).end(); it1++)
		 {
			  for(list<HypergraphInfo>::iterator it2 = (*itTmp).begin(); it2!=(*itTmp).end(); it2++)
			  {
					 if(it1->id!=it2->link)
						 continue;

					 set<IDTYPE> nodes1=it1->hg.getVertices();
					 set<IDTYPE> nodes2=it2->hg.getVertices();

           for(set<IDTYPE>::iterator itnode1 = nodes1.begin(); itnode1!=nodes1.end(); itnode1++)
           {
						 if(nodes2.count(*itnode1)!=0)
							 if(it1->hg.getEqLabels(*itnode1).size()!=0)
								 it2->hg.setEqLabels(*itnode1, it1->hg.getEqLabels(*itnode1));
           }
        }
		 }
	 }
}

/** is the graph with id at the level linked from some graph at the level + 1 */
bool EquivalenceConstructor::isLinked(unsigned int id, unsigned int level)
{
		if(level>=m_list.size()-1)
			return false;

    // last not empty level in m_list
	  for(list<HypergraphInfo>::iterator it = m_list[level+1].begin(); it!= m_list[level+1].end(); it++)
	  {
			 if(it->link==id)
				 return true;
	  }

	  return false;
}
