
#include "ExampleChecker.h"

#include "GXLWriter.h"

#include <stdexcept>

ExampleChecker::ExampleChecker(Petrigraph & pg, vector<IDTYPE> & trans)
{
	  this->pg = pg;

	  this->trans = trans;
	  
	  counter = 0;
}

ExampleChecker::~ExampleChecker()
{
}

vector<list<HypergraphInfo> > ExampleChecker::check(unsigned int * trans_num, bool * result)
{
  try
  {
	*trans_num=0;
	
	GraphGrammar gg = pg.getGraphGrammar();

	set<IDTYPE> v = pg.getHypergraph().getVertices();

  m_list.resize(1);
	// start graph
  m_list[0].push_back(HypergraphInfo(pg.getGraphGrammar().getStartGraph(), counter, 0));

  counter++;

  unsigned int count = 0;

  while(1)
  {
    if(trans.size()==0)
    {
			*result = true;
			return m_list;
		}	

    // resize
  	m_list.resize(count+2);
    		
    checkLevel(count);

	  if(!pg.getPTNet().fireTransition(trans[count]))
			throw runtime_error("");

		count++;

		if( (count==trans.size()) && (m_list[count].size()!=0) )
		{
			*result = true;
			return m_list;
		}

		if(m_list[count].size()==0)
		{
			*trans_num=count;
			*result = false;
  		return m_list;
		}
  }
  
	return m_list;
	}
	catch(exception e)
	{
		throw e;
	}
}

/** check [M_i] set */
void ExampleChecker::checkLevel(int level)
{
	GraphGrammar ggr = pg.getGraphGrammar();

  try
  {
    map<IDTYPE, IDTYPE> mergedEdges;

		// graph(M)
		Hypergraph hmarking = pg.getMarkedHypergraph(&mergedEdges);

		MatchHandler mhandler;

		IDTYPE id_rule = pg.getPTNet().getTransition(trans[level]).getAssociatedRule();

		TransformationRule r = ggr.getRule(id_rule);

		// L
		Hypergraph hleft = r.getLeft();

		list<HypergraphInfo>::iterator it = m_list[level].begin();

		// must have at least one element
		assert(it!=m_list[level].end());

		int size_m = m_list[level].size();
		
		for(int i = 0; i<size_m; i++)
		{
			// matches from L to G
			list<map<IDTYPE, IDTYPE> > matchesLG = it->hg.findAllMatchings(hleft, mhandler, false);

			// matches from G to graph(M)
			list<map<IDTYPE, IDTYPE> > matchesGgM = hmarking.findAllMatchings((it->hg), mhandler, true);

      // choosen matches L -> G
      list<map<IDTYPE, IDTYPE> > matches;

			// check commuting
			for(list<map<IDTYPE, IDTYPE> >::iterator lg = matchesLG.begin(); lg!=matchesLG.end(); lg++)
			{
				bool commutes = false;
				for(list<map<IDTYPE, IDTYPE> >::iterator gm = matchesGgM.begin(); gm!=matchesGgM.end(); gm++)
				{
						map<IDTYPE, IDTYPE> & mu_mapping = pg.getMu(trans[level]);

						if(checkCommuting((*lg), (*gm), mu_mapping, mergedEdges))
					  {
							commutes=true;
							break;
					  }
				}

				if(commutes)
				{
					matches.push_back(*lg);
					commutes=false;
				}
			}

			// apply all matchings
			for(list<map<IDTYPE, IDTYPE> >::iterator it_match = matches.begin(); it_match!=matches.end(); it_match++)
			{
				Hypergraph h = (it->hg).applyRule(r.getLeft(), r.getRight(), (*it_match));

				m_list[level+1].push_back(HypergraphInfo(h, counter, it->id));
				it->matchesLG = matches;
				counter++;
			}

			// if it is the last not empty element of m_list,
			// then initialize matchesGgM 
//			if(m_list[level+1].size()==0)
//			{
				for(list<map<IDTYPE, IDTYPE> >::iterator itGgMList = matchesGgM.begin();
									 itGgMList!=matchesGgM.end(); itGgMList++)
				{
					for(map<IDTYPE, IDTYPE>::iterator itGgM = (*itGgMList).begin();
										 itGgM!=(*itGgMList).end(); itGgM++)
					{
						if(mergedEdges.find(itGgM->second)!=mergedEdges.end())
							itGgM->second=mergedEdges[itGgM->second];
					}
				}
				it->matchesGgM = matchesGgM;
//			}


		it++;
		}
	}
	catch(exception e)
	{
		cout << "error by checking the level" << endl;
		throw e;
	}

}

/** check if the diagram commutes */
bool ExampleChecker::checkCommuting(map<IDTYPE, IDTYPE> ab,
																		map<IDTYPE, IDTYPE> bc,
																		map<IDTYPE, IDTYPE> ac,
																		map<IDTYPE, IDTYPE> mergedEdges)
{
   for(map<IDTYPE, IDTYPE>::iterator it = ab.begin(); it!=ab.end(); it++)
	 {
		 IDTYPE vertedge;
		 map<IDTYPE, IDTYPE>::iterator itFind = mergedEdges.find(bc[it->second]);

		 if(itFind==mergedEdges.end())
			 vertedge = bc[it->second];
		 else
		   vertedge = itFind->second;
		 
		 if(vertedge!=ac[it->first])
		 	 return false;
	 }

	 return true;
}




