#include "GXLWriter.h"
#include "PTNet.h"
#include "Hypergraph.h"
#include "Petrigraph.h"
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <cstdlib>
#include <stdexcept>
#include <list>
#include <set>


using namespace std;

  GXLWriter::GXLWriter(char *pfilename)
  {
    int n = strlen(pfilename);
    
    filename = new char[n+1];
    strcpy(filename,pfilename);
    doc = xmlNewDoc((const xmlChar *)"1.0");
    xmlNodePtr root = xmlNewNode(NULL,(const xmlChar *)"gxl");
    xmlDocSetRootElement(doc,root);

  }


  GXLWriter::~GXLWriter()
  {
    xmlFreeDoc(doc);
    delete filename;
  }


  void GXLWriter::addGraph(Petrigraph& pg, char const* id)
  {
    xmlNodePtr root = xmlDocGetRootElement(doc);

    xmlNodePtr graph = xmlNewChild(root,NULL,(const xmlChar *)"graph",NULL);
    xmlSetProp(graph,(const xmlChar *)"id",(const xmlChar *)id);
    xmlSetProp(graph,(const xmlChar *)"hypergraph",(const xmlChar *)"true");
    xmlSetProp(graph,(const xmlChar *)"edgemode",(const xmlChar *)"undirected");

    set<IDTYPE> vertices = pg.getHypergraph().getVertices();
    set<IDTYPE> edges = pg.getHypergraph().getEdges();
    list<PTNetTransition> transitions = pg.getPTNet().getTransitions();

    // obtain graph Grammar
    GraphGrammar & gr = pg.getGraphGrammar();

    // obtain startgraph
    Hypergraph & start_graph = gr.getStartGraph();

    // store vertices:
    for(set<IDTYPE>::iterator i = vertices.begin();
        i!=vertices.end();
        i++)
    {
      char nID[256];
      IDTYPE2xmlID(*i,nID);
      xmlNodePtr v = xmlNewChild(graph,NULL,(const xmlChar *)"node",NULL);
      xmlSetProp(v,(const xmlChar *)"id",(const xmlChar *)nID);
      xmlNodePtr attr = xmlNewChild(v,NULL,(const xmlChar *)"attr",NULL);
      xmlSetProp(attr,(const xmlChar *)"name",(const xmlChar *)"vertex");

      // iota_mapping
      if(mapping)
      {
      xmlNodePtr iotaAttr = xmlNewChild(v,NULL,(const xmlChar *)"attr",NULL);
      xmlSetProp(iotaAttr, (const xmlChar *)"name",(const xmlChar *)"iota_mapping");

      string s;
      list<IDTYPE> lst = pg.iota_minus_1(*i);
      for(list<IDTYPE>::iterator it=lst.begin(); it!=lst.end(); it++)
      {
          bool found;
          s += start_graph.getIdLabel(*it, &found);
					list<IDTYPE>::iterator ittmp = it;
					if(++ittmp!=lst.end())
						s += ";";
			}
      xmlNewChild(iotaAttr,NULL,(const xmlChar *)"string",(const xmlChar *)s.c_str());
      }

    }

    //obtain initial marking
    PTMarking initial_marking = pg.getPTNet().getInitialMarking();

    // store edges:
    map<IDTYPE,xmlNodePtr> id2xmlNode;
    for(set<IDTYPE>::iterator i = edges.begin();
        i!=edges.end();
        i++)
    {
      char rID[256];
      IDTYPE2xmlID(*i,rID);
      xmlNodePtr rel = xmlNewChild(graph,NULL,(const xmlChar *)"rel",NULL);
      xmlSetProp(rel,(const xmlChar *)"id",(const xmlChar *)rID);

      xmlNodePtr attr = xmlNewChild(rel,NULL,(const xmlChar *)"attr",NULL);
      xmlSetProp(attr,(const xmlChar *)"name",(const xmlChar *)"label");
      xmlNewChild(attr,NULL,(const xmlChar *)"string",(const xmlChar *)(pg.getHypergraph().getLabel(*i).c_str()));

      xmlNodePtr markingAttr = xmlNewChild(rel, NULL, (const xmlChar*) "attr", NULL);
      xmlSetProp(markingAttr, (const xmlChar *)"name", (const xmlChar *)"initial_marking");
      char buf[256];
      sprintf(buf, "%d", initial_marking[*i]);
      xmlNewChild(markingAttr, NULL, (const xmlChar *)"int", (const xmlChar*) buf);

      // jota_mapping
      if(mapping)
      {
			xmlNodePtr iotaAttr = xmlNewChild(rel,NULL,(const xmlChar *)"attr",NULL);
      xmlSetProp(iotaAttr, (const xmlChar *)"name",(const xmlChar *)"iota_mapping");

      string s;
      list<IDTYPE> lst = pg.iota_minus_1(*i);
      bool found;
      for(list<IDTYPE>::iterator it=lst.begin(); it!=lst.end(); it++)
      {
          s += start_graph.getIdLabel(*it, &found);
					list<IDTYPE>::iterator ittmp = it;
					if(++ittmp!=lst.end())
						s += ";";
			}
      xmlNewChild(iotaAttr,NULL,(const xmlChar *)"string",(const xmlChar *)s.c_str());
      }

      list<IDTYPE> c = pg.getHypergraph().getConnectedVertices(*i);
      // append connections
      int count = 0;
      for(list<IDTYPE>::iterator j = c.begin();
          j!=c.end();
          j++)
      {
        char nID[256];
        char chCount[256];
        IDTYPE2xmlID(*j,nID);
        sprintf(chCount,"%i",count);
        xmlNodePtr relend = xmlNewChild(rel,NULL,(const xmlChar *)"relend",NULL);
        xmlSetProp(relend,(const xmlChar *)"target",(const xmlChar*)nID);
        xmlSetProp(relend,(const xmlChar *)"role",(const xmlChar*)"vertex");
        xmlSetProp(relend,(const xmlChar *)"startorder",(const xmlChar *)chCount);
        count++;
      }
      id2xmlNode[*i] = rel;
    }
    // store transitions and attach to rels
    for(list<PTNetTransition>::iterator i = transitions.begin();
        i!=transitions.end();
        i++)
    {
      char tID[256];
      IDTYPE2xmlID((*i).getID(),tID);

      //char rID[256];
      //IDTYPE2xmlID((*i).getAssociatedRule(),rID);
      const char* rID = (pg.getGraphGrammar().getRule( (*i).getAssociatedRule() ).getName().c_str());

      xmlNodePtr t = xmlNewChild(graph,NULL,(const xmlChar *)"node",NULL);
      xmlSetProp(t,(const xmlChar *)"id",(const xmlChar *)tID);
      xmlNodePtr attr = xmlNewChild(t,NULL,(const xmlChar *)"attr",NULL);
      xmlSetProp(attr,(const xmlChar *)"name",(const xmlChar *)"transition");
      xmlNodePtr attr2 = xmlNewChild(t,NULL,(const xmlChar *)"attr",NULL);
      xmlSetProp(attr2,(const xmlChar *)"name",(const xmlChar *)"rule");
      xmlNewChild(attr2,NULL,(const xmlChar *)"string",(const xmlChar *)rID);

      // mu_mapping
      if(mapping)
      {
      xmlNodePtr muAttr = xmlNewChild(t,NULL,(const xmlChar *)"attr",NULL);
      xmlSetProp(muAttr, (const xmlChar *)"name",(const xmlChar *)"mu_mapping");

      Hypergraph & hleft = pg.getGraphGrammar().getRule( (*i).getAssociatedRule() ).getLeft();
      Hypergraph & hright = pg.getGraphGrammar().getRule( (*i).getAssociatedRule() ).getRight();
                  
      string s = "";
      map<IDTYPE, IDTYPE> mel = pg.getMuElements((*i).getID());
      bool found;
      for(map<IDTYPE, IDTYPE>::iterator it=mel.begin(); it!=mel.end(); it++)
      {
					string tmp = hleft.getIdLabel(it->first, &found);
					if(tmp!="")
						s += tmp;
					else
						s += hright.getIdLabel(it->first, &found);
					s +=",";
					
					char buf[15];
					strcpy(buf, "");
					sprintf(buf, "%u", it->second);
					s += buf;
					map<IDTYPE, IDTYPE>::iterator ittmp = it;
					if(++ittmp!=mel.end())
						s += ";";
			}

      xmlNewChild(muAttr,NULL,(const xmlChar *)"string",(const xmlChar *)s.c_str());
      }
        
      set<IDTYPE> pre = (*i).getPreset();
      for(set<IDTYPE>::iterator j = pre.begin();
          j!=pre.end();
          j++)
      {
        char chWeight[256];
        sprintf(chWeight,"%i",(*i).getPreWeight(*j));

        xmlNodePtr rel = id2xmlNode[*j];
        xmlNodePtr relend = xmlNewChild(rel,NULL,(const xmlChar *)"relend",NULL);
        xmlSetProp(relend,(const xmlChar *)"target",(const xmlChar*)tID);
        xmlSetProp(relend,(const xmlChar *)"role",(const xmlChar*)"preset");
        xmlNodePtr attr3 = xmlNewChild(relend,NULL,(const xmlChar *)"attr",NULL);
        xmlSetProp(attr3,(const xmlChar *)"name",(const xmlChar *)"weight");
        xmlNewChild(attr3,NULL,(const xmlChar *)"int",(const xmlChar *)chWeight);
      }
      set<IDTYPE> post = (*i).getPostset();
      for(set<IDTYPE>::iterator j = post.begin();
          j!=post.end();
          j++)
      {
        char chWeight[256];
        sprintf(chWeight,"%i",(*i).getPostWeight(*j));

        xmlNodePtr rel = id2xmlNode[*j];
        xmlNodePtr relend = xmlNewChild(rel,NULL,(const xmlChar *)"relend",NULL);
        xmlSetProp(relend,(const xmlChar *)"target",(const xmlChar*)tID);
        xmlSetProp(relend,(const xmlChar *)"role",(const xmlChar*)"postset");
        xmlNodePtr attr3 = xmlNewChild(relend,NULL,(const xmlChar *)"attr",NULL);
        xmlSetProp(attr3,(const xmlChar *)"name",(const xmlChar *)"weight");
        xmlNewChild(attr3,NULL,(const xmlChar *)"int",(const xmlChar *)chWeight);
      }
    }
  }


  void GXLWriter::write()
  {
    if(-1 == xmlSaveFormatFile(filename,doc,true))
      throw runtime_error("Error writing XML file");
  }
