#include "Hypergraph.h"
#include "Globals.h"

#include <assert.h>

using namespace std;


  //IDTYPE Hypergraph::lastUsedID(0);

  Hypergraph::Hypergraph()
  {
    
  }

  Hypergraph::Hypergraph(const Hypergraph& h)
    :    vertices(h.vertices),
	 edges(h.edges),
	 connectedV(h.connectedV),
	 connectedE(h.connectedE),
	 labels(h.labels),
	 idLabels(h.idLabels),
	 eq_labels(h.eq_labels)
  {
    /*    vertices = h.vertices;
    edges = set<IDTYPE>(h.edges);
    connectedV = map<IDTYPE,list<IDTYPE> >(h.connectedV);
    connectedE = map<IDTYPE,multiset<IDTYPE> >(h.connectedE);
    for(map<IDTYPE,LABELTYPE>::const_iterator i = h.labels.begin();
        i!=h.labels.end();
        i++)
    {
       labels[(*i).first] = LABELTYPE((*i).second);
    }

    */
  }

  Hypergraph::~Hypergraph()
  {

  }  

  /** Adds a new vertex to the hypergraph. The ID of the created
      vertex is returned. */
  IDTYPE Hypergraph::addVertex()
  {
    IDTYPE newv = createNewID();
    vertices.insert(newv);
    connectedE.insert(pair<IDTYPE,multiset<IDTYPE> >(newv,multiset<IDTYPE>()));
    return newv;
  }

  IDTYPE Hypergraph::addVertex(IDTYPE id)
  {
    if(0==vertices.count(id))
      {
	
	vertices.insert(id);
	connectedE.insert(pair<IDTYPE,multiset<IDTYPE> >(id,multiset<IDTYPE>()));
      }
    return id;
  }
  
  
  /** Adds a new hyperedge to the hypergraph.
      connection is a list of vertices in the hypergraph.
      label is a label for the new edge.
      The ID of the new edge is returned.
  */

  IDTYPE Hypergraph::addEdge(list<IDTYPE> connection, LABELTYPE label)
  {
    IDTYPE newe = createNewID();
    edges.insert(newe);
    connectedV.insert(pair<IDTYPE,list<IDTYPE> >(newe,connection));
    labels.insert(pair<IDTYPE,LABELTYPE>(newe,LABELTYPE(label)));

    for(list<IDTYPE>::iterator i = connection.begin();
	i != connection.end();
	i++)
      {
	assert(1==vertices.count(*i)); // connection must not contain other IDs than vertices of this hypergraph
	// find multiset of edges connected to vertex *i
	map<IDTYPE,multiset<IDTYPE> >::iterator j = connectedE.find(*i);
	(*j).second.insert(newe);
      }
    return newe;
  }

  IDTYPE Hypergraph::addEdge(list<IDTYPE> connection, LABELTYPE label, IDTYPE id)
  {
    if(id==3047)
    {
			int i;
			i++;
    }

    if(0==edges.count(id))
      {
	edges.insert(id);
	connectedV.insert(pair<IDTYPE,list<IDTYPE> >(id,connection));
	labels.insert(pair<IDTYPE,LABELTYPE>(id,LABELTYPE(label)));
	
	for(list<IDTYPE>::iterator i = connection.begin();
	    i != connection.end();
	    i++)
	  {
	    assert(1==vertices.count(*i)); // connection must not contain other IDs than vertices of this hypergraph
	    // find multiset of edges connected to vertex *i
	    map<IDTYPE,multiset<IDTYPE> >::iterator j = connectedE.find(*i);
	    (*j).second.insert(id);
	  }
      }
    return id;
  }

  void Hypergraph::uniongraphs(const Hypergraph& h)
  {
    set<IDTYPE> existed;

    // add vertices of h to this:
    for(set<IDTYPE>::iterator i = h.vertices.begin();
        i != h.vertices.end(); i++)
      {
        if(0 == vertices.count(*i) )
					vertices.insert(*i);
				else
					existed.insert(*i);
      }
      
    // add edges and labels of h to this:
    for(set<IDTYPE>::const_iterator j = h.edges.begin();
	j != h.edges.end();
	j++)
      {
	// assert edge sets are disjunct
	assert (0 == edges.count(*j));
	edges.insert(*j);
	labels.insert(*(h.labels.find(*j)));
	idLabels.insert(*(h.idLabels.find(*j)));
      }
      
    // add connectivity mapping
    // for edges
    for(map<IDTYPE,list<IDTYPE> >::const_iterator k = h.connectedV.begin();
	k != h.connectedV.end();
	k++)
      {
	connectedV.insert(*k);
      }
      
    // connectedE  
    for(map<IDTYPE,multiset<IDTYPE> >::const_iterator l = h.connectedE.begin();
	l != h.connectedE.end();
	l++)
      {
				if(existed.find(l->first)!=existed.end())
				{
					map<IDTYPE,multiset<IDTYPE> >:: iterator c = connectedE.find(l->first);
					if(c!=connectedE.end())
					{
						for(multiset<IDTYPE>::iterator mit = (l->second).begin(); mit!=(l->second).end(); mit++)
							(c->second).insert(*mit);
					}
				}
				else	
					connectedE.insert(*l);
      }
  }

  LABELTYPE Hypergraph::getLabel(IDTYPE  edge) const
  {
    assert(1 == edges.count(edge));

    return (*(labels.find(edge))).second;
  }
  
  map<IDTYPE,LABELTYPE> Hypergraph::getLabels()
  {
    return labels;
  }

  std::set<IDTYPE> Hypergraph::getVertices() 
  {
    return vertices;
  }

   /** set id label */
   void Hypergraph::setIdLabel(IDTYPE id, std::string ls)
   {
		 idLabels[id]=ls;
	 }

	 /** get id label */
	 std::string Hypergraph::getIdLabel(IDTYPE id, bool * found)
	 {
		 if(idLabels.find(id)!=idLabels.end())
		 {
				 *found = true;
				 return idLabels[id];
		 }
		 else
		 {
				 *found = false; 
				 return "";
		 }	 
	 }

	 /** get ID for the label */
	 IDTYPE Hypergraph::getLabelId(std::string s, bool * found)
	 {
		 for(std::map<IDTYPE, std::string>::iterator it=idLabels.begin(); it!=idLabels.end(); it++)
		 {
       if(it->second==s)
       {
				*found = true; 
				return it->first;
			 }
				
		 }
		 *found = false;
		 return 0;
	 }
  
  std::set<IDTYPE> Hypergraph::getEdges()
  {
    return edges;
  }
  
  list<IDTYPE> Hypergraph::getConnectedVertices(IDTYPE edge)
  {
    assert(1 == edges.count(edge));

    map<IDTYPE, list<IDTYPE> >::iterator it = connectedV.find(edge);
      
    if(it!=connectedV.end())
			return it->second;
		else
			return emptyIDTYPE;
  }

  multiset<IDTYPE> Hypergraph::getConnectedEdges(IDTYPE vertex) const
  {
    assert (1 == vertices.count(vertex));

    return (*(connectedE.find(vertex))).second;
  }

  bool Hypergraph::setDepth(IDTYPE id, long depth, VERTEDGE v)
  {
      if(v==EDGE)
      	depthmapEdges[id] = depth;
      else
				depthmapVertices[id] = depth;
				 
			return true;
  }
  bool Hypergraph::getDepth(IDTYPE id, long * pdepth, VERTEDGE v)
  {
		*pdepth = 0;		
    // edge
    if(v==EDGE)
    {
    std::set<IDTYPE>::iterator it = edges.find(id);

		if(it==edges.end())
			return false;
		else
		{
			*pdepth = depthmapEdges[id];
			return true;
		}
		}
		// vertice
		else
		{
    std::set<IDTYPE>::iterator it = vertices.find(id);

		if(it==vertices.end())
			return false;
		else
		{
			*pdepth = depthmapVertices[id];
			return true;
		}			
		}
  }

  void Hypergraph::initDepth()
  {
   for(std::set<IDTYPE>::iterator it = edges.begin(); it != edges.end(); it++)
			depthmapEdges[(*it)] = 0;
			
		for(std::set<IDTYPE>::iterator it = vertices.begin(); it != vertices.end(); it++)
			depthmapVertices[(*it)] = 0;
  }

  std::map<IDTYPE, long> Hypergraph::getDepthTable(VERTEDGE v)
  {
		if(v==EDGE)
			return depthmapEdges;
		else
			return depthmapVertices;
  }

  void Hypergraph::setDepthTable(std::map<IDTYPE, long> & depthtable, VERTEDGE v)
  {
    if(v==EDGE)
    {
			depthmapEdges.empty();
			depthmapEdges = depthtable;
		}
		else
		{
			depthmapVertices.empty();
			depthmapVertices = depthtable;
		}
  }
  
     map<IDTYPE,IDTYPE> Hypergraph::findMatching(Hypergraph& h, MatchHandler& handler)
  {
    // initialise datastructures
    std::stack<IDTYPE> unmapped_edges;

    stack<set<IDTYPE>::iterator> mappings;

    for(set<IDTYPE>::iterator i = h.edges.begin();
	i != h.edges.end();
	i++)
     unmapped_edges.push(*i);

    map<IDTYPE,IDTYPE> phi_v;
    map<IDTYPE,IDTYPE> phi_e;

    map<IDTYPE,IDTYPE> result;
    if(findNextMatching(h,mappings,unmapped_edges,phi_v,phi_e,handler)) // Does a matching exist?
      {
	// return union of phi_v and phi_e
	for(map<IDTYPE,IDTYPE>::iterator v = phi_v.begin();
	    v != phi_v.end();
	    v++)
	  result.insert(*v);
	for(map<IDTYPE,IDTYPE>::iterator e = phi_e.begin();
	    e != phi_e.end();
	    e++)
	  result.insert(*e);
	return result;
      }
      else // no matching found
      {
	// findNextMatching should have cleared phi and mappings
	assert(0 == phi_v.size());
	assert(0 == phi_e.size());
	assert(0 == mappings.size());
	assert(unmapped_edges.size() == h.edges.size());
	return result; // return an empty map
      }
  }


  // private
  bool Hypergraph::findNextMatching(Hypergraph &h,
				    stack<set<IDTYPE>::iterator> &mappings,
				    stack<IDTYPE> &unmapped_edges,
				    map<IDTYPE,IDTYPE> &phi_v,
				    map<IDTYPE,IDTYPE> &phi_e,
				    MatchHandler& matchhandler)
  {
    // a matching was found?
    if(0 == unmapped_edges.size())
      {
	map<IDTYPE,IDTYPE> thematching;
	for(map<IDTYPE,IDTYPE>::iterator v = phi_v.begin();
	    v != phi_v.end();
	    v++)
	  thematching.insert(*v);
	for(map<IDTYPE,IDTYPE>::iterator e = phi_e.begin();
	    e != phi_e.end();
	    e++)
	  thematching.insert(*e);	
	return matchhandler.useThisMatch(thematching);
      }
    else // the matching is not yet complete
      {
	// edges and vertices in *this are named with prefix g,
	// edges and vertices in h are named with prefix h
	// get the next edge to be mapped
	IDTYPE hedge = unmapped_edges.top();
  	unmapped_edges.pop();
	// iterator on mappings stack iterates through all edges of g
	// to find an appropriate mapping
	mappings.push(edges.begin());
	// label of hedge:
	LABELTYPE hedgelabel = h.getLabel(hedge);
	// all vertices connected to the edge to be mapped
	list<IDTYPE> hcV = h.getConnectedVertices(hedge);
	// iterate mappings.top() throung all edges of g
	while(mappings.top() != edges.end())
	  {
	    IDTYPE gedge = *(mappings.top());
	    // label of gedge:
	    LABELTYPE gedgelabel = getLabel(gedge);
 	    // all vertices connected to the mapping target
	    list<IDTYPE> gcV = getConnectedVertices(gedge);
	    // the label and cardinality of matching edges must be the same
      if((hedgelabel != gedgelabel) || (gcV.size() != hcV.size()))
	      {
		(mappings.top())++; // try another edge
		continue; 
	      }
	    // store all new vertice mappings to delete them later
	    list<IDTYPE> newMappedVertices;
	    bool vertexMatchingFailed(false);
	    

	    list<IDTYPE>::iterator hv = hcV.begin();
	    list<IDTYPE>::iterator gv = gcV.begin();
	    
	    while((!vertexMatchingFailed) && hv != hcV.end())
	    {
	      
		map<IDTYPE,IDTYPE>::iterator phi_v_it = phi_v.find(*hv);
		if(phi_v.end() == phi_v_it) // vertex is not yet mapped
		  {
		    newMappedVertices.push_back(*hv);
		    phi_v.insert(pair<IDTYPE,IDTYPE>(*hv,*gv));
		  }
		else //vertex is already mapped
		  if((*phi_v_it).second != *gv) //mismatch, abort
		    vertexMatchingFailed = true;
		gv++;
		hv++;
	    }
	    // now all vertices are mapped, or vertexMatchingFailed==true
	    if(!vertexMatchingFailed)
	      {
		phi_e.insert(pair<IDTYPE,IDTYPE>(hedge,gedge));
		if(findNextMatching(h,mappings,unmapped_edges,phi_v,phi_e,matchhandler)) return true;
		phi_e.erase(hedge);
	      }
	    // undo all changes in phi_v
	    for(list<IDTYPE>::iterator v = newMappedVertices.begin();
		v != newMappedVertices.end();
		v++)
	      phi_v.erase(*v);
	    (mappings.top())++; // try next edge mapping
	  
	  }
	unmapped_edges.push(hedge);
	mappings.pop();
	
	return false; // nothing found
      }
    
  }


  ostream& operator<<(ostream& os, Hypergraph& hg)
  {
    os << "vertices: ";
    for(set<IDTYPE>::iterator i = hg.vertices.begin();
	i != hg.vertices.end();
	i++)
      os << (*i) << " ";
    os << "  edges: ";
    for(set<IDTYPE>::iterator i = hg.edges.begin();
	i!= hg.edges.end();
	i++)
      {
	os << (*i) << " <" << hg.getLabel(*i) << ">(";
	map<IDTYPE,list<IDTYPE> >::iterator j = hg.connectedV.find(*i);
	for(list<IDTYPE>::iterator k = (*j).second.begin();
	    k != (*j).second.end();
	    k++)
	  os << (*k) << " ";
	os << ") ";
      }
    return os;
  }


/** find all matchings */
list<map<IDTYPE, IDTYPE> > Hypergraph::findAllMatchings(Hypergraph& h, MatchHandler& handler, bool surject)
{
    stack<IDTYPE> unmapped_edges;

		// initialise datastructures
    stack<set<IDTYPE>::iterator> mappings;

    for(set<IDTYPE>::iterator i = h.edges.begin();
	i != h.edges.end();
	i++)
     unmapped_edges.push(*i);

    map<IDTYPE,IDTYPE> phi_v;
    map<IDTYPE,IDTYPE> phi_e;

    list<map<IDTYPE, IDTYPE> > found_matches;

    findAllMatchingsNext(h,mappings,unmapped_edges,phi_v,phi_e,found_matches, handler, surject);

    // findNextMatching should have cleared phi and mappings
    assert(0 == phi_v.size());
    assert(0 == phi_e.size());
    assert(0 == mappings.size());
    assert(unmapped_edges.size() == h.edges.size());

    return found_matches;
}
/** find next matching for all matchings search */
void Hypergraph::findAllMatchingsNext(Hypergraph& h,
																			stack<set<IDTYPE>::iterator>& mappings,
																			stack<IDTYPE>& unmapped_edges,
																			map<IDTYPE, IDTYPE>& phi_v,
																			map<IDTYPE, IDTYPE>& phi_e,
																			std::list<map<IDTYPE, IDTYPE> >& found_matches,
																			MatchHandler& handler,
																			bool surject)
{
	  // a matching was found?
    if(0 == unmapped_edges.size())
      {
    // surjective
    bool surject_checked = false;
		if(surject)
		{
  		 set<IDTYPE> ev;
 	  	 for(map<IDTYPE,IDTYPE>::iterator v = phi_v.begin(); v != phi_v.end(); v++)
             if(ev.count(v->second)==0)
								 ev.insert(v->second);
       for(map<IDTYPE,IDTYPE>::iterator e = phi_e.begin(); e != phi_e.end(); e++)
             if(ev.count(e->second)==0)
								 ev.insert(e->second);

			 if( ev.size()==(vertices.size()+edges.size()) )
		     surject_checked = true;
		   else
		     surject_checked = false;  
		}
		else
			surject_checked = true;

		if(surject_checked)
		{
		map<IDTYPE,IDTYPE> thematching;
	  for(map<IDTYPE,IDTYPE>::iterator v = phi_v.begin();
	    v != phi_v.end();
	    v++)
	  thematching.insert(*v);
	  for(map<IDTYPE,IDTYPE>::iterator e = phi_e.begin();
	    e != phi_e.end();
	    e++)
	  thematching.insert(*e);
			
		if(handler.useThisMatch(thematching))
				 found_matches.push_back(thematching);
		} 
      }
    else // the matching is not yet complete
      {
	// edges and vertices in *this are named with prefix g,
	// edges and vertices in h are named with prefix h
	// get the next edge to be mapped
	IDTYPE hedge = unmapped_edges.top();
  	unmapped_edges.pop();
	// iterator on mappings stack iterates through all edges of g
	// to find an appropriate mapping
	mappings.push(edges.begin());
	// label of hedge:
	LABELTYPE hedgelabel = h.getLabel(hedge);
	// all vertices connected to the edge to be mapped
	list<IDTYPE> hcV = h.getConnectedVertices(hedge);
	// iterate mappings.top() throung all edges of g
	while(mappings.top() != edges.end())
	  {
	    IDTYPE gedge = *(mappings.top());
	    // label of gedge:
	    LABELTYPE gedgelabel = getLabel(gedge);
 	    // all vertices connected to the mapping target
	    list<IDTYPE> gcV = getConnectedVertices(gedge);
	    // the label and cardinality of matching edges must be the same
      if((hedgelabel != gedgelabel) || (gcV.size() != hcV.size()))
	      {
		(mappings.top())++; // try another edge
		continue;
	      }
 	    // store all new vertice mappings to delete them later
	    list<IDTYPE> newMappedVertices;
	    bool vertexMatchingFailed(false);


	    list<IDTYPE>::iterator hv = hcV.begin();
	    list<IDTYPE>::iterator gv = gcV.begin();

	    while((!vertexMatchingFailed) && hv != hcV.end())
	    {

		map<IDTYPE,IDTYPE>::iterator phi_v_it = phi_v.find(*hv);
		if(phi_v.end() == phi_v_it) // vertex is not yet mapped
		  {
		    newMappedVertices.push_back(*hv);
		    phi_v.insert(pair<IDTYPE,IDTYPE>(*hv,*gv));
		  }
		else //vertex is already mapped
		  if((*phi_v_it).second != *gv) //mismatch, abort
		    vertexMatchingFailed = true;
		gv++;
		hv++;
	    }
	    // now all vertices are mapped, or vertexMatchingFailed==true
	    if(!vertexMatchingFailed)
	      {
		phi_e.insert(pair<IDTYPE,IDTYPE>(hedge,gedge));
		
		findAllMatchingsNext(h,mappings,unmapped_edges,phi_v,phi_e,found_matches,handler,surject);
		
		phi_e.erase(hedge);
	      }
	    // undo all changes in phi_v
	    for(list<IDTYPE>::iterator v = newMappedVertices.begin();
		v != newMappedVertices.end();
		v++)
	      phi_v.erase(*v);
	    (mappings.top())++; // try next edge mapping

	  }
	unmapped_edges.push(hedge);
	mappings.pop();

	return;
      }
}

/** apply the rule */
Hypergraph Hypergraph::applyRule(Hypergraph& left, Hypergraph& right, map<IDTYPE, IDTYPE> & matching)
{
		Hypergraph hg(*this);

    set<IDTYPE> vertices_left = left.getVertices();
    set<IDTYPE> vertices_right = right.getVertices();

    set<IDTYPE> vertices_to_insert;

    set<IDTYPE> remained_nodes;

    for(set<IDTYPE>::iterator v = vertices_right.begin(); v != vertices_right.end(); v++)
    {
			if(vertices_left.find(*v)==vertices_left.end())
				vertices_to_insert.insert(*v);
			else
				remained_nodes.insert(*v);
		}

    // map to the inserted ids
		map<IDTYPE, IDTYPE> inserted;
		
    // add vertices from r to hypergraph
    for(set<IDTYPE>::iterator v = vertices_to_insert.begin(); v != vertices_to_insert.end(); v++)
    {
			inserted[*v] = createNewID();
    	hg.addVertex(inserted[*v]);
      bool found;
      hg.setIdLabel(inserted[*v], right.getIdLabel(*v, &found));
      assert(found);
    }

    // map to the remained ids
    map<IDTYPE, IDTYPE> remained;
    
    // add remained nodes in r to the map
    for(set<IDTYPE>::iterator v = remained_nodes.begin(); v!=remained_nodes.end(); v++)
    {
			map<IDTYPE, IDTYPE>::iterator m = matching.find(*v);
			assert(m!=matching.end());
			remained[*v]=m->second;
    }

    //edges
    set<IDTYPE> edges_left = left.getEdges();
    set<IDTYPE> edges_right = right.getEdges();

    // delete L edges
    for(set<IDTYPE>::iterator it=edges_left.begin(); it!=edges_left.end(); it++)
			  hg.deleteEdge(matching[*it]);

		// change R Ids
		Hypergraph R_changed(right);

 		R_changed.changeID(inserted, remained);

		// union
		hg.uniongraphs(R_changed);

		return hg;
}

/** delete the edge */
void Hypergraph::deleteEdge(IDTYPE id)
{
	  edges.erase(id);
	  connectedV.erase(id);

	  labels.erase(id);
	  idLabels.erase(id);

	  for(map<IDTYPE,multiset<IDTYPE> >::iterator it=connectedE.begin(); it!=connectedE.end(); it++)
	  {
			(it->second).erase(id);
	  }
}

/** change the Ids accordingly to the maps
inserted are the inserted nodes
changed are the remained nodes
all edges should obtain another ids
R and Hypergraph will be united
*/
void Hypergraph::changeID(map<IDTYPE, IDTYPE> & inserted, map<IDTYPE, IDTYPE> & changed)
{
	 // inserted	
	 for(map<IDTYPE, IDTYPE>::iterator it = inserted.begin(); it!=inserted.end(); it++)
	 {
		  vertices.erase(it->first);
		  vertices.insert(it->second);
	 }		

   // changed
   for(map<IDTYPE, IDTYPE>::iterator it = changed.begin(); it!=changed.end(); it++)
	 {
      vertices.erase(it->first);
      vertices.insert(it->second);
   }

   map<IDTYPE, IDTYPE> edges_map;
   
   // edges
   for(set<IDTYPE>::iterator it =edges.begin(); it!=edges.end(); it++)
   {
  	 IDTYPE id = createNewID();
		 edges_map[*it]=id;
   }

   // change edges
   for(map<IDTYPE, IDTYPE>::iterator it = edges_map.begin(); it!=edges_map.end(); it++)
   {
		 edges.erase(it->first);
 		 edges.insert(it->second);

		 // labels
		 labels[it->second] = labels[it->first];
		 labels.erase(it->first);

		 // id labels
		 idLabels[it->second] = idLabels[it->first];
		 idLabels.erase(it->first);

		 // connected vertices
		 for(list<IDTYPE>::iterator itV = connectedV[it->first].begin(); itV!=connectedV[it->first].end(); itV++)
		 {
	 		 //	node
	     IDTYPE node_id = 0;

	     map<IDTYPE, IDTYPE>::iterator ins = inserted.find(*itV);
	     if(ins==inserted.end())
	     {
				 map<IDTYPE, IDTYPE>::iterator ch = changed.find(*itV);
				 assert(ch!=changed.end());
						node_id = ch->second;
			 }
			 else
			   node_id = ins->second;

			 connectedV[it->second].push_back(node_id);
		 }

		 connectedV.erase(it->first);

	 }

	 // connectedE

	 map<IDTYPE,std::multiset<IDTYPE> > connectedE_changed;
	 
	 for(map<IDTYPE,std::multiset<IDTYPE> >::iterator it = connectedE.begin(); it!=connectedE.end(); it++)
	 {
		 // edges
		 multiset<IDTYPE> edges_ms;
		 for(multiset<IDTYPE>::iterator it1 = (it->second).begin(); it1!=(it->second).end(); it1++)
					edges_ms.insert(edges_map[*it1]);
					
		 //	node
		 IDTYPE node_id = 0;

		 map<IDTYPE, IDTYPE>::iterator ins = inserted.find(it->first);
		 if(ins==inserted.end())
		 {
			  map<IDTYPE, IDTYPE>::iterator ch = changed.find(it->first);
			  assert(ch!=changed.end());
					node_id = ch->second;
		 }
		 else
			 node_id = ins->second;

		 connectedE_changed[node_id] = edges_ms;
	 }

	 connectedE.clear();

	 connectedE = connectedE_changed;
}

/** set the equivalence class label */
bool Hypergraph::setEqLabel(IDTYPE id_node, unsigned int label)
{
  for(list<unsigned int>::iterator it = eq_labels[id_node].begin(); it!=eq_labels[id_node].end(); it++)
		if( *it == label )
			return false;
		
	eq_labels[id_node].push_back(label);

	return true;
}

/** get equivalence class fothe node */
list<unsigned int> & Hypergraph::getEqLabels(IDTYPE id_node)
{
   if(eq_labels[id_node].size()!=0)
		 return eq_labels[id_node];
	 else
		 return empty;
}

/** get the number of set labels */
unsigned int Hypergraph::getEqCount()
{
	return eq_labels.size();
}

/** set labels */
void Hypergraph::setEqLabels(IDTYPE id_node, list<unsigned int> labels)
{
	 eq_labels[id_node]=labels;
}
