#include "Petrigraph.h"
#include "aunfold.h"

using namespace std;

  Petrigraph::Petrigraph(Hypergraph& hg,
			 PTNet& ptn,
			 GraphGrammar& ts,
			 map<IDTYPE,map<IDTYPE,IDTYPE> > mu,
			 map<IDTYPE,IDTYPE> iota)
    : hypergraph(hg),
      ptnet(ptn),
      transformationsystem(ts),
      mu_mapping(mu),
      iota_mapping(iota)

  {
  }

  Petrigraph::Petrigraph(const Petrigraph& ptg)
    :hypergraph(ptg.hypergraph),
     ptnet(ptg.ptnet),
     transformationsystem(ptg.transformationsystem),
     mu_mapping(ptg.mu_mapping),
     iota_mapping(ptg.iota_mapping)
  {
  }

  Petrigraph::Petrigraph()
  {
		realFolding = true;
  }


  Hypergraph& Petrigraph::getHypergraph() 
  {
    return hypergraph;
  }

  PTNet& Petrigraph::getPTNet()
  {
    return ptnet;
  }

  GraphGrammar& Petrigraph::getGraphGrammar()
  {
    return transformationsystem;
  }


  void Petrigraph::fold(TransformationRule& r,
    map<IDTYPE,IDTYPE> phi2,
    map<IDTYPE,IDTYPE> phi)
  {
    Relation rel(this);
    Hypergraph& left = r.getLeft();
    set<IDTYPE> edges_left = left.getEdges();
    
    for(set<IDTYPE>::iterator e = edges_left.begin();
	e!=edges_left.end();
	e++)
      {
	IDTYPE phi2_e = (*(phi2.find(*e))).second;
	IDTYPE phi_e =  (*(phi.find(*e))).second;
	rel.addEdgeRel(phi2_e,phi_e);
      }

    merge(rel);

    if(red)
    ReduceTransitions();
  }

  void Petrigraph::unfold(TransformationRule& r,
	      map<IDTYPE,IDTYPE> phi)
  {

    // first, add P(t,r) to this Petrigraph as defined between Def. 9 and 10
    Hypergraph& left = r.getLeft();
    Hypergraph& right = r.getRight();
    set<IDTYPE> vertices_left = left.getVertices();
    set<IDTYPE> vertices_right = right.getVertices();
    map<IDTYPE,IDTYPE> mu_t;

    // add vertices from r to this petri graph:
    for(set<IDTYPE>::iterator v = vertices_left.begin();
	v != vertices_left.end();
	v++)
      {
	hypergraph.addVertex(*v);
	mu_t[*v] = *v;

	long dep;
	hypergraph.getDepth(phi[*v], &dep, VERTICE);
	hypergraph.setDepth(*v, dep, VERTICE);
      }

    for(set<IDTYPE>::iterator v = vertices_right.begin();
	v != vertices_right.end();
	v++)
      {      
	hypergraph.addVertex(*v);
	mu_t[*v] = *v;
      }

    set<IDTYPE> edges_left = left.getEdges();
    set<IDTYPE> edges_right = right.getEdges();
    map<IDTYPE,unsigned> preweight;
    map<IDTYPE,unsigned> postweight;
    // initialize preweight and postweight with 0's
    set<IDTYPE> places = ptnet.getPlaces();
    for(set<IDTYPE>::iterator i = places.begin();
	i!=places.end();
	i++)
      {
	preweight[*i] = 0;
	postweight[*i] = 0;
      }

    long max_depth = 0;
    
    for(set<IDTYPE>::iterator e = edges_left.begin();
	e!=edges_left.end();
	e++)
      { 
	hypergraph.addEdge(left.getConnectedVertices(*e),left.getLabel(*e),*e);
	preweight[*e] = 1;
	ptnet.addPlace(*e);
	mu_t[*e] = *e;

	long dep;
	hypergraph.getDepth(phi[*e], &dep, EDGE);
	hypergraph.setDepth(*e, dep, EDGE);

	//depth
	if(dep > max_depth)
			max_depth = dep;
      }

    for(set<IDTYPE>::iterator e = edges_right.begin();
	e!=edges_right.end();
	e++)
      {
	hypergraph.addEdge(right.getConnectedVertices(*e),right.getLabel(*e),*e);
	postweight[*e] = 1;
	ptnet.addPlace(*e);
	mu_t[*e] = *e;

	hypergraph.setDepth(*e, max_depth+1, EDGE);
      }

    for(set<IDTYPE>::iterator v = vertices_right.begin();
	v != vertices_right.end();
	v++)
		hypergraph.setDepth(*v, max_depth+1, VERTICE);

    // create the new transition:
    PTNetTransition t(edges_left,edges_right,preweight,postweight,r.getID());
    ptnet.addTransition(t);
    mu_mapping[t.getID()] = mu_t;
    
    Relation rel(this);

    for(set<IDTYPE>::iterator e = edges_left.begin();
	e!=edges_left.end();
	e++)
      {
	IDTYPE phi_e = (*(phi.find(*e))).second;
	rel.addEdgeRel(phi_e,*e);
      }

    merge(rel);

    if(red)
    ReduceTransitions();

    // set the depth in the rule to 0
    left.initDepth();
    right.initDepth();
    
   }

  void Petrigraph::ReduceTransitions()
  {
		bool merged = true;

		while(merged)
		{

		merged = false;
		
		list<PTNetTransition> transList;
		transList = ptnet.getTransitions();

    for(list<PTNetTransition>::iterator it= transList.begin(); it!= transList.end(); it++)
    {
				// rule
			  IDTYPE ruleID = (*it).getAssociatedRule();
			  // preset
			  set<IDTYPE> preset = (*it).getPreset();

			  // relation for merging of transitions
			  Relation rel(this);
			  
			  list<PTNetTransition>::iterator tmp = it;
			  tmp++;
        // second transition
			  for(list<PTNetTransition>::iterator it1 = tmp; it1!= transList.end(); it1++)
			  {
					 // rule
					 IDTYPE rule1ID = (*it1).getAssociatedRule();
					 if(ruleID != rule1ID)
						 continue;

					 // preset
					 set<IDTYPE> preset1 = (*it1).getPreset();

					 // compare presets
					 if( !(preset == preset1) )
						 continue;

					 bool weigths = true;
					 	 
					 // preset weigths
					 for(set<IDTYPE>::iterator itSet = preset.begin(); itSet != preset.end(); itSet++)
					 {
						  unsigned n1 = (*it).getPreWeight(*itSet);
 						  unsigned n2 = (*it1).getPreWeight(*itSet);

					    if(n1 != n2)
					    {
								weigths=false;
								break;
							}
					 }

					 if(!weigths)
						 continue;

				  // merge 
					IDTYPE t = it->getID();
					IDTYPE t1 = it1->getID();
					
					rel.addTransitionRel(t, t1);

					merged = true;
					
				}  // second transition

        if(merged)
        {
					merge(rel);
          break;
        }
	
    } // first

    } // while
}
   
  IDTYPE Petrigraph::mu(IDTYPE transition, IDTYPE graphel)
  {
    map<IDTYPE,IDTYPE>& matching = (*(mu_mapping.find(transition))).second;
    return (*(matching.find(graphel))).second;
  }

  IDTYPE Petrigraph::iota(IDTYPE graphel)
  {
    return iota_mapping[graphel];
  }

  ostream& operator<<(ostream& os, Petrigraph& pg)
  {
    os << "Petrigraph: \n" << pg.hypergraph << "\n" << pg.ptnet << "\n\n\n";
    return os;
  }

  /** returns mu map for the given transition */
  map<IDTYPE, IDTYPE> Petrigraph::getMuElements(IDTYPE transition)
  {
		map<IDTYPE,IDTYPE>& matching = (*(mu_mapping.find(transition))).second;

		return matching;
	}
   
  //private:
  void Petrigraph::merge(Relation& r)
  {
    set<IDTYPE> V = getHypergraph().getVertices();
    set<IDTYPE> E = getHypergraph().getEdges();
    list<PTNetTransition> T = getPTNet().getTransitions();

    set<IDTYPE> nV; // new vertex set
    set<IDTYPE> nE; // new edge set
    list<PTNetTransition> nT; // new transitions
    set<IDTYPE> handledTransitions; // to ensure each transition is inserted only once
    map<IDTYPE,list<IDTYPE> > nc; // new connection function
    map<IDTYPE,LABELTYPE> nl; // new labeling function
    PTMarking nMarking;  // new initial marking of the ptnet
    map<IDTYPE,map<IDTYPE,IDTYPE> > nMu; // new mu mapping
    map<IDTYPE,IDTYPE> nIota; // new iota mapping
    
    // get a mapping from this ptgraph to the new IDs of the
    // equivalence classes
    map<IDTYPE,IDTYPE> equivrel = r.getExtendingEquivRelation();

    map<IDTYPE, IDTYPE> eq_copy = equivrel;

    if(realFolding)
    {
    // change the matches in checked Foldings   
    for(list<CheckedFoldingsStructure>::iterator it = SecondMatchHandler::checkedFoldings.begin();
					it!=SecondMatchHandler::checkedFoldings.end(); it++)
		{
			  for(map<IDTYPE, IDTYPE>::iterator itM=it->match.begin(); itM!=it->match.end(); itM++)
			  {
          if(eq_copy[itM->second]!=0)
						itM->second = eq_copy[itM->second];
				}

				for(map<IDTYPE, IDTYPE>::iterator itM=it->matchFirst.begin(); itM!=it->matchFirst.end(); itM++)
				{
					if(eq_copy[itM->second]!=0)
						itM->second = eq_copy[itM->second];
				}
		}
		}
		else
			realFolding=true;
    
    // merge    
    if(debug)
      {
	cout << "Equivrel: " << equivrel << "\n";
      }
    
    
    for(map<IDTYPE,IDTYPE>::iterator i = equivrel.begin();
	i!= equivrel.end();
	i++)
      {
	// initialize new vertices:
	if(1==V.count((*i).first))
	  nV.insert((*i).second);
	
	// initialize new edges:
	else if(1==E.count((*i).first))
	  {
	    nE.insert((*i).second);
	    list<IDTYPE> connect = getHypergraph().getConnectedVertices((*i).first);
	    list<IDTYPE> nconnect;
	    for(list<IDTYPE>::iterator j = connect.begin();
		j!=connect.end();
		j++)
	      nconnect.push_back(equivrel[*j]);
	    nc.insert(pair<IDTYPE,list<IDTYPE> >((*i).second,nconnect));
	    LABELTYPE newlabel = getHypergraph().getLabel((*i).first);
	    nl.insert(pair<IDTYPE,LABELTYPE>((*i).second,newlabel));
	    // nMarking will be updated later
	    nMarking.insert(pair<IDTYPE,unsigned>((*i).second,0));
	    
	  }
	// initialize new transitions:
	else
	  {
	    // Transition is new?
	    if(0==handledTransitions.count((*i).second))
	      {
		// get transition of the existing Petrigraph
		PTNetTransition& transition = getPTNet().getTransition((*i).first);
		// calculate preset, postset and weights of the new transition.
		// this is a multiset join where the "weight" refers to the
		// multiplicity of the set items in the multiset.
		set<IDTYPE> preset = transition.getPreset();
		set<IDTYPE> nPreset;
		map<IDTYPE,unsigned> nPreWeight;
		for(set<IDTYPE>::iterator j = preset.begin();
		    j!= preset.end();
		    j++)
		  {
		    if(0==nPreset.count(equivrel[*j]))
		      {
		    
			nPreset.insert(equivrel[*j]);
			nPreWeight.insert(pair<IDTYPE,unsigned>
					  (equivrel[*j], transition.getPreWeight(*j)));
		      }

		    else // there is already a weight, we have to add
		      {
			unsigned oldweight = nPreWeight[equivrel[*j]];
			nPreWeight[equivrel[*j]] = oldweight + transition.getPreWeight(*j);
		      }
		    
		  }
	    
		set<IDTYPE> postset = transition.getPostset();
		set<IDTYPE> nPostset;
		map<IDTYPE,unsigned> nPostWeight;
		for(set<IDTYPE>::iterator j = postset.begin();
		    j!= postset.end();
		    j++)
		  {
		    if(0==nPostset.count(equivrel[*j]))
		      {
		    
			nPostset.insert(equivrel[*j]);
			nPostWeight.insert(pair<IDTYPE,unsigned>
					   (equivrel[*j], transition.getPostWeight(*j)));
		      }

		    else // there is already a weight, we have to add
		      {
			unsigned oldweight = nPostWeight[equivrel[*j]];
			nPostWeight[equivrel[*j]] = oldweight + transition.getPostWeight(*j);
		      }
		  }
	    
		// create new transition. The new ID stored in equivrel is discarded and replaced by a new one.
		PTNetTransition ntransition(nPreset,nPostset,nPreWeight,nPostWeight,transition.getAssociatedRule());
		nT.push_back(ntransition);
		// calculate new mu(transition)(x):
		map<IDTYPE,IDTYPE> mu_transition = mu_mapping[transition.getID()];
		map<IDTYPE,IDTYPE> mu_nTransition;
		for(map<IDTYPE,IDTYPE>::iterator j = mu_transition.begin();
		    j!=mu_transition.end();
		    j++)
		  {
		    mu_nTransition[(*j).first] = equivrel[(*j).second];
		  }
		if(debug)
		  {
		    cout << "mu_nTransition= " << mu_nTransition << "\n";
		  }
		
		nMu.insert(pair<IDTYPE,map<IDTYPE,IDTYPE> >(ntransition.getID(),
							    mu_nTransition));
		handledTransitions.insert((*i).second);
	      }
	  }
	
    
      }
    
    // calculate new iota mapping and initial marking
    for(map<IDTYPE,IDTYPE>::iterator j = iota_mapping.begin();
	j!=iota_mapping.end();
	j++)
      {
	IDTYPE newGraphElement = equivrel[(*j).second];
	
	nIota.insert(pair<IDTYPE,IDTYPE>((*j).first,newGraphElement));
	if(1==E.count((*j).second)) // this is an edge
	  nMarking[newGraphElement] = 1 + nMarking[newGraphElement];
      }

    // Now replace old petrigraph by the merged one.

    // first copy the depth table
    std::map<IDTYPE, long> depthtableEdge = hypergraph.getDepthTable(EDGE);
		std::map<IDTYPE, long> depthtableVertice = hypergraph.getDepthTable(VERTICE);
					
    hypergraph = Hypergraph();

    hypergraph.setDepthTable(depthtableEdge, EDGE);
    hypergraph.setDepthTable(depthtableVertice, VERTICE);
    
    for(set<IDTYPE>::iterator i = nV.begin();
	i!=nV.end();
	i++)
      hypergraph.addVertex(*i);

    for(set<IDTYPE>::iterator i=nE.begin();
	i!=nE.end();
	i++)
      hypergraph.addEdge(nc[*i],nl[*i],*i);

    ptnet = PTNet(nE,nT,nMarking);
    mu_mapping = nMu;
    iota_mapping = nIota;
    // merging complete.
    
  }

void Petrigraph::setIota(std::map<IDTYPE,IDTYPE> iota_map)
{
	iota_mapping = iota_map;
}

list<IDTYPE> Petrigraph::iota_minus_1(IDTYPE e_v)
{
    list<IDTYPE> lst;

    for(map<IDTYPE,IDTYPE>::iterator it = iota_mapping.begin();
									it!=iota_mapping.end(); it++)
				if(it->second == e_v)
					lst.push_back(it->first);

		return lst;
}

/** get hypergraph corresponding to the current marking of PN */
Hypergraph Petrigraph::getMarkedHypergraph(map<IDTYPE, IDTYPE> * mergedEdges)
{
	 Hypergraph hg;

	 PTMarking current_marking = ptnet.getCurrentMarking();

	 list<IDTYPE> edges;

	 for(PTMarking::iterator it=current_marking.begin(); it!=current_marking.end(); it++)
	 {
		  for(unsigned int i=0; i<(it->second); i++)
				edges.push_back(it->first);
	 }

	 for(list<IDTYPE>::iterator it_edges = edges.begin(); it_edges!=edges.end(); it_edges++)
	 {
		   list<IDTYPE> vertices = hypergraph.getConnectedVertices(*it_edges);
		   
		   for(list<IDTYPE>::iterator it_vertices = vertices.begin(); it_vertices != vertices.end(); it_vertices++)
				 hg.addVertex(*it_vertices);

			 string label = hypergraph.getLabel(*it_edges);

			 IDTYPE edge = hg.addEdge(vertices, label);

       if(mergedEdges!=NULL)
				 (*mergedEdges)[edge] = (*it_edges);
	 }

	 return hg;
}

/** get mu mapping */
map<IDTYPE, IDTYPE> & Petrigraph::getMu(IDTYPE trans_id)
{
	return mu_mapping[trans_id];
}


/** No descriptions */
void Petrigraph::unrealFolding()
{
	realFolding=false;
}
