#ifndef PETRIGRAPH_H
#define PETRIGRAPH_H

#include "Hypergraph.h"
#include "TransformationRule.h"
#include "GraphGrammar.h"
#include "PTNet.h"
#include "Relation.h"

#include <map>

  /**Definition of a petri graph representation.
   * a relation as defined in Definition 7 is included.
   * 
   */
  class Petrigraph
    {
    public:
      Petrigraph(Hypergraph& hg,
		 PTNet& ptn,
		 GraphGrammar& ts,
		 std::map<IDTYPE,std::map<IDTYPE,IDTYPE> > mu,
		 std::map<IDTYPE,IDTYPE> iota);
      
      Petrigraph(const Petrigraph& ptg);

      Petrigraph();

      Hypergraph& getHypergraph();
      PTNet& getPTNet();
      GraphGrammar& getGraphGrammar();

      /** Petrigraph folding (Definition 9)*/
      void fold(TransformationRule& r,
		std::map<IDTYPE,IDTYPE> phi2,
		std::map<IDTYPE,IDTYPE> phi);

      /**  Petrigraph unfolding (definition 10) */
      void unfold(TransformationRule& r,
		  std::map<IDTYPE,IDTYPE> phi);

		  void ReduceTransitions();
      
      /** mu mapping as defined in definition (6):
	  maps a transition to a matching of the lefthand/righthand side
	  of the associated transformation rule in the hypergraph.
      */
      IDTYPE mu(IDTYPE transition, IDTYPE graphel);

      /** iota mapping as defined in definition (6):
	  maps vertices and edges of the start graph of the graph grammar
	  to the vertices and edges of the petri graph. */
      IDTYPE iota(IDTYPE graphel);

      /** set iota values */
      void setIota(std::map<IDTYPE,IDTYPE> iota_map);

      list<IDTYPE> iota_minus_1(IDTYPE e_v);

      std::map<IDTYPE, IDTYPE> getMuElements(IDTYPE transition);

			Hypergraph getMarkedHypergraph(map<IDTYPE, IDTYPE> * mergedEdges = NULL);

			map<IDTYPE, IDTYPE> & getMu(IDTYPE trans_id);
  /** No descriptions */
  void unrealFolding();

      friend std::ostream& operator<<(std::ostream& os, Petrigraph& pg);

    private:
      // merge Petrigraph as described in Definition 8
      void merge(Relation& r);

      bool realFolding;
      
      Hypergraph hypergraph;
      PTNet ptnet;
      GraphGrammar transformationsystem;
      // transition_id -> (transformationruleelement_id -> hypergraphelement_id)
      std::map<IDTYPE,std::map<IDTYPE,IDTYPE> > mu_mapping;
      std::map<IDTYPE,IDTYPE> iota_mapping;
    };


     struct HypergraphInfo
     {
			  HypergraphInfo(Hypergraph hg,
											 unsigned int id,
											 unsigned int link)
			  {
					 this->hg = hg;
					 this->id = id;
					 this->link = link;
			  }
			  HypergraphInfo()
			  {
			  }
			  Hypergraph hg;
			  unsigned int id;
			  unsigned int link;
			  list<map<IDTYPE, IDTYPE> > matchesLG;
			  // is initialized only for last not empty set [M_k]
			  list<map<IDTYPE, IDTYPE> > matchesGgM;
		 };
		 


#endif
