
#include "SampleStorage.h"
#include "EquivalenceConstructor.h"

#include <stdexcept>
#include <string.h>
#include <fstream>
#include <iostream>

SampleStorage::SampleStorage(char * samplefile)
{
	
}

SampleStorage::~SampleStorage()
{
	
}

/** add m_list to the samples */
void SampleStorage::addSample(vector<list<HypergraphInfo> > & hi, vector<string> & trans)
{
   Sample sam(hi, trans);
	 m_list_storage.push_back(sam);
}

/** check petrigraph */
bool SampleStorage::checkPetrigraph(Petrigraph & pg)
{
  try
  {
  
	for(list<Sample>::iterator it = m_list_storage.begin(); it!=m_list_storage.end(); it++)
	{
	 // reachable markings
	 list<PTMarking> markings;
	 
	 // set initial marking in Petrigraph
	 pg.getPTNet().setCurrentMarking(pg.getPTNet().getInitialMarking());
	 
	 // start with initial
	 markings.push_back(pg.getPTNet().getInitialMarking());
	 
   // transitions
	 vector<string>::iterator itTrans = (*it).trans.begin();
	 
   // tree
	 for(vector<list<HypergraphInfo> >::iterator itVec = (*it).tree.begin(); itVec!=(*it).tree.end(); itVec++)
	 {
		 // hypergraphs
		 for(list<HypergraphInfo>::iterator itH = (*itVec).begin(); itH!=(*itVec).end(); itH++)
		 {
		   // markings
			 for(list<PTMarking>::iterator itM = markings.begin(); itM!=markings.end(); itM++)
			 {	
				 pg.getPTNet().setCurrentMarking(*itM);
				 Hypergraph hgM = pg.getMarkedHypergraph();

				 // matches
         MatchHandler mh;
				 list<map<IDTYPE, IDTYPE> > matches;

				 if(automor)
					 matches = hgM.findAllMatchings(itH->hg, mh, true);
				 else
				 {
					 map<IDTYPE, IDTYPE> m = hgM.findMatching(itH->hg, mh);
					 matches.push_back(m);
				 }

				 for(list<map<IDTYPE, IDTYPE> >::iterator itMat = matches.begin(); itMat!=matches.end(); itMat++)
				 {
					 if(!checkLabels(hgM, itH->hg, *itMat))
						 return false;
				 }
			 }
		 } 

		 // end of transitions
		 if(itTrans == (*it).trans.end())
			 break;

		 // calculate new markings
		 list<PTMarking> markings_tmp;

		 // markings
		 for(list<PTMarking>::iterator itM = markings.begin(); itM!=markings.end(); itM++)
		 {
			 pg.getPTNet().setCurrentMarking(*itM);

			 bool found;

			 IDTYPE id_rule = pg.getGraphGrammar().findRule(*itTrans, &found);

			 if(!found)
			 {
			  throw runtime_error("internal error");
			 }
			 
       // possible transitions
			 set<IDTYPE> posSet = pg.getPTNet().getPossibleIds(id_rule);

			 for(set<IDTYPE>::iterator its = posSet.begin(); its!=posSet.end(); its++)
			 {
				 pg.getPTNet().fireTransition(*its);
				 markings_tmp.push_back(pg.getPTNet().getCurrentMarking());
				 pg.getPTNet().setCurrentMarking(*itM);
			 }
		 }

		 if(markings_tmp.size()==0)
			 break;

		 markings.empty();
		 markings = markings_tmp; 

		 // transition list
		 itTrans++;
	 }

	}
	
	return true;
	
	}
	catch(exception e)
	{
		cout << "Error by checking the sample" << endl;
		throw e;
	}
}

/** check if some nodes having the same labels are matches in one */
bool SampleStorage::checkLabels(Hypergraph & hg1, Hypergraph & hg2, map<IDTYPE, IDTYPE> & match)
{
	try
	{
		// h2 -> h1
		set<IDTYPE> nodes1 = hg1.getVertices();
		set<IDTYPE> nodes2 = hg2.getVertices();

		for(set<IDTYPE>::iterator it=nodes1.begin(); it!=nodes1.end(); it++)
		{
       set<IDTYPE> selected_nodes;
       
			 for(map<IDTYPE, IDTYPE>::iterator itM = match.begin(); itM!=match.end(); itM++)
			 {
				  if(itM->second == *it)
            selected_nodes.insert(itM->first);
			 }

			 if(selected_nodes.size()>1)
			 {
				 multiset<unsigned int> mlabels;
				 
				 // check labels
         for(set<IDTYPE>::iterator itS = selected_nodes.begin(); itS!=selected_nodes.end(); itS++)
         {
					  // labels
					  list<unsigned int> labels = hg2.getEqLabels(*itS);

					  // add to multiset
					  for(list<unsigned int>::iterator itL = labels.begin(); itL!=labels.end(); itL++)
							mlabels.insert(*itL);
				 }

         // check
         for(multiset<unsigned int>::iterator itMultiset = mlabels.begin(); itMultiset != mlabels.end(); itMultiset++)
         	 if(mlabels.count(*itMultiset) > 1)
						 return false;
         
			 }
		}

		return true;
	}
	catch(exception e)
	{
		throw e;
	}
}

void SampleStorage::WriteXML(const char *filename)
{
    xmlDocPtr doc;

    doc = xmlNewDoc((const xmlChar *)"1.0");
    xmlNodePtr root = xmlNewNode(NULL,(const xmlChar *)"m_list_storage");
    xmlDocSetRootElement(doc,root);


	xmlNodePtr storage_xml = xmlDocGetRootElement(doc);


	for(list<Sample>::iterator it = m_list_storage.begin(); it!=m_list_storage.end(); it++)
	{ 
		WriteSample(storage_xml, *it);
	}

   if(-1 == xmlSaveFormatFile(filename,doc,true))
      throw runtime_error("Error writing XML file");
                  
}

void SampleStorage::ReadXML(const char *filename)
{
try
{
	using namespace std;
	ifstream file;
	file.open(filename);
	if(file.fail())
		return;
	else
	{
	m_list_storage.clear();
	xmlDocPtr doc = xmlParseFile(filename);
	if (doc == NULL )
	{
		fprintf(stderr,"Document not parsed successfully. \n");
		xmlFreeDoc(doc);
	}	
	
	xmlNodePtr storage = xmlDocGetRootElement(doc);
	if (storage == NULL)
		throw runtime_error("Empty xml document");
	
		
	if (xmlStrcasecmp(storage->name, (const xmlChar*) "m_list_storage"))
		throw runtime_error("Root node 'm_list_storage' expected");
	

	xmlNodePtr sample_in_list = GetChildrenNode(storage,(char*)"sample");
	
	Sample sample;
	sample.tree.clear();
	sample.trans.clear();
	
	while(sample_in_list != NULL)
	{	
		ReadSample(sample_in_list, sample);
		m_list_storage.push_back(sample);
		sample.tree.clear();
		sample.trans.clear();
		sample_in_list = GetNextChild(sample_in_list, (char*)"sample");
	}
	}
	file.close();
}
catch(exception e)
{throw e;}
}
 
void SampleStorage::WriteSample(xmlNodePtr list_sample, Sample &sample)
{
	EquivalenceConstructor* eq = new EquivalenceConstructor ();
	xmlNodePtr sample_in_list = xmlNewChild(list_sample, NULL, (const xmlChar *)"sample", NULL);
	
	xmlNodePtr sample_tree = xmlNewChild(sample_in_list, NULL, (const xmlChar *)"tree", NULL);
	eq->Write_m_list(sample_tree, sample.tree);
	
	xmlNodePtr sample_trans = xmlNewChild(sample_in_list, NULL, (const xmlChar *)"trans", NULL);
	for(vector<string>::iterator it = sample.trans.begin(); it != sample.trans.end(); it++)
	{
		xmlNewChild(sample_trans, NULL, (const xmlChar *)"string_in_vector", (const xmlChar *) (*it).c_str());
	}
	
	delete eq;
}

void SampleStorage::ReadSample(xmlNodePtr list_sample, Sample &sample)
{
try
{
	EquivalenceConstructor* eq = new EquivalenceConstructor ();
	
	xmlNodePtr sample_tree = GetChildrenNode(list_sample, (char*)"tree");
	
	if(sample_tree != NULL)
	{
		eq->Read_m_list(sample_tree, sample.tree);
	}
	
	xmlNodePtr sample_trans = NULL;
	sample_trans = GetChildrenNode(list_sample, (char*)"trans");
	xmlNodePtr cur = NULL;
	
	if(sample_trans != NULL) 
	{
		string str;
		xmlNodePtr string_in_trans = GetChildrenNode(sample_trans, (char*)"string_in_vector");
	
		while(string_in_trans != NULL)
		{
			cur = string_in_trans->xmlChildrenNode;
			if(cur != NULL)
			{
				str = (char*)cur->content;
				sample.trans.push_back(str);
				cur = NULL;
			}
			
			string_in_trans = GetNextChild(string_in_trans, (char*) "string_in_vector");
		}
	}
	delete eq;
}
catch(exception e)
{throw e;}
}

xmlNodePtr SampleStorage:: GetChildrenNode(xmlNodePtr parent, char* name_child)
{
	xmlNodePtr child = parent->xmlChildrenNode;
	while((child != NULL) && (xmlStrcasecmp(child->name, (const xmlChar*)name_child)))
	{
		child = child->next;
	}
	
	return child;
	
		
}

xmlNodePtr SampleStorage:: GetNextChild(xmlNodePtr child1, char* name_child)
{
	
	xmlNodePtr child2 = child1->next;
	while((child2 != NULL) && (xmlStrcasecmp(child2->name, (const xmlChar*)name_child)))
	{
		child2 = child2->next;
	}
	
	return child2;
	
	
}


