#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <stdlib.h>

#include <fstream>

#include "GXLReader.h"
#include "GTXLReader.h"
#include "TransReader.h"

#include "aunfold.h"
#include "GXLWriter.h"
#include "ExampleChecker.h"
#include "EquivalenceConstructor.h"
#include "SampleStorage.h"

using namespace std;

int main(int argc, char *argv[])
{
	const char* usage = "Usage: unfolding [-q] [-d] [-t] [-nc] [-depth=k | -nfold=k] [-step=n] \
	approxFile, grammarFile, transitionsFile, outputFile, sampleFile  \n\
	\nOptions:\n\n\
	-q quiet      quiet mode, disable verbose output\n\
	-d debug      debug mode, show many debugging information\n\
	-t trace      include each step of calculation into output file\n\
	-nc           do not check covering information\n\
	-nr           without irredundancy condition \n\
	-depth=k      build over-upproximation C(k) \n\
	-step=n       make only n steps of the algorithm\n\
	-nfold=k      no folding steps with maximal depth of unfolding \n\
	-timeout=n    timeout in sek\n\n";

	if(argc<6)
	{
		cout << usage << endl;
		return 1;
	}

	int i;
	
  for(i=1; i<argc; i++)
    {
      if(!strcmp(argv[i],"-q")) // quiet mode
	quiet = true;
      else if(!strcmp(argv[i],"-d")) // debug output
	debug = true;
      else if(!strcmp(argv[i],"-t")) // trace
	trace = true;
      else if(!strcmp(argv[i],"-nc")) // do not check covering
	ncov = true;
			else if(!strcmp(argv[i],"-nr")) // do not check irredundancy condition
	red = false;
			else if(!strcmp(argv[i],"-nmapping")) // without mapping functions
	mapping = false;
			else if(strstr(argv[i],"-depth=")==argv[i])
			{
        char buffer[10];
				strcpy(buffer, argv[i]+7);
				depth = atoi(buffer);

				if(depth<0)
				{
					cout << "depth must be some positive number or 0" << "\n";
					return 1;
				}
			}
			else if(strstr(argv[i], "-nfold=")==argv[i]) // no folding steps
			{
				if(depth>0)
				{
					cout << "-depth and -nfold cannot be defined together";
					return 1;
				}

				nfold = true;

        char buffer[10];
				strcpy(buffer, argv[i]+7);
				unfdepth = atoi(buffer);

				if(unfdepth<=0)
				{
					cout << "unfolding depth must be some positive number" << "\n";
					return 1;
				}

			}
			else if(strstr(argv[i],"-step=")==argv[i])
			{
        char buffer[10];
				strcpy(buffer, argv[i]+6);
				maxstep = atoi(buffer);

				if(maxstep<=0)
				{
					cout << "step must be some positive number" << "\n";
					return 1;
				}
			}
			else if(strstr(argv[i],"-timeout=")==argv[i])
			{
        char buffer[10];
				strcpy(buffer, argv[i]+9);
				timeout = atoi(buffer);

				if(timeout<0)
				{
					cout << "timeout must be some positive number or 0" << "\n";
					return 1;
				}
			}
      else if(*argv[i] == '-')
	{
	  cout << "Unknown option: " << argv[i] << "\n" << usage;
	  return 1;
	}
      else
	break;
    }
  if(i > argc-5)
    {
      cout << usage;
      return 1;
    }

  char* infileapprox = argv[i];
	char* infilegramm = argv[i+1];
	char* transfile = argv[i+2];
	char* outfile = argv[i+3];
	char* samplefile = argv[i+4];

	automor = true;

  SampleStorage samplestorage(samplefile);

  bool exception_c = false;
  	
  vector<IDTYPE> translist;

  vector<string> transstrings;

  Petrigraph pg;

   GraphGrammar ts;
  
	try
	{	
		GTXLReader grReader(infilegramm);
		ts = grReader.read();
					
    		GXLReader reader(infileapprox);
		reader.setGraphGrammar(ts);
		long addID;
		pg = reader.read(&addID);
		
		TransReader tr(transfile);
		tr.read(addID);
		translist = tr.getIDs();
		transstrings = tr.getStrings();

//		GXLWriter writer(outfile);
//		writer.addGraph(pg, "result");
//		writer.write();
	}
	catch(exception e)
  {
     cerr << "\n" << "Error while reading input files\n";
     exception_c=true;
  }
  
  // [M_0] ... [M_k+1] =? empty
  vector<list<HypergraphInfo> > vlh;

  // checking
  if(!exception_c)
  {
  try
  {
		ExampleChecker echecker(pg, translist);

		unsigned int trans_num;
    bool result;
		
		vlh = echecker.check(&trans_num, &result);

		if(result)
		{
			cout << "The example is real \n";
			ofstream echt("echt");
			return EXIT_SUCCESS;
		}	
		else
		{
			cout << "The example is spurious \n" << "Transition number " << trans_num << ", id: ";
			cout << translist[trans_num-1] << ", name: " << transstrings[trans_num-1];
			cout << ",\n" << "could not be fired \n";
		}
  }
  catch(exception e)
  {
     cerr << "\n Error while checking the example \n Possible reason: two edges with the same label in the left-hand side of the rule";
     exception_c=true;
  }
  }

  // make equivalence classes
  if(!exception_c)
  {
		try
		{
			EquivalenceConstructor econstructor(pg, translist, vlh);

			vlh.empty();

			vlh = econstructor.construct();
						
		}
		catch(exception e)
		{
			 cerr << "\n Error by building of equivalence classes \n Possible reason: two edges with the same label in the left-hand side of the rule";
       exception_c=true;
		}
  }

  GXLWriter writer(outfile);

  if(!exception_c)
  {
		try
		{
			samplestorage.ReadXML(samplefile);
			samplestorage.addSample(vlh, transstrings);
		}
		catch(exception e)
		{
			cerr << "\n Error by reading of sample file \n";
			exception_c = true;
		}
  }
  
  bool btimeout = false;
  ofstream of;

  if(!exception_c)
  {
		try
		{
	    double timer = double(clock()) / CLOCKS_PER_SEC;
			Petrigraph pg (approximatedUnfolding(ts,writer, samplestorage, timer, &btimeout));
			writer.addGraph(pg,"result");
	    double timeused = double(clock()) / CLOCKS_PER_SEC - timer;
	    cout<< "time used: " << timeused << " sek \n";

      of.open("result", ofstream::out | ofstream::app);

	    of << "time used: " << timeused << " sek \n";
	    	    
      if(!btimeout)
      {
      // petrigraph size
	    cout << "edges: " << pg.getHypergraph().getEdges().size();
	    cout << ", vertices: " << pg.getHypergraph().getVertices().size();
	    cout << ", transitions: " << pg.getPTNet().getTransitions().size() << endl;

      // petrigraph size
	    of << "edges: " << pg.getHypergraph().getEdges().size();
	    of << ", vertices: " << pg.getHypergraph().getVertices().size();
	    of << ", transitions: " << pg.getPTNet().getTransitions().size() << "\n";
	    }
	    else
				of << "timeout \n";

		}
		catch (exception& e)
		{
			cerr << "\n Error by unfolding \n" << "Possible reason: bad or multiple Ids in GTS" << "\n";
			exception_c=true;
		}
	}	

  if(!exception_c)
  {
		try
    {
			if(!btimeout)
			{
				writer.write();
	
	samplestorage.WriteXML(samplefile);
		
      }
    }
    catch(exception e)
    {
      cerr << "Error while writing output files\n";
    }
  }

  return EXIT_SUCCESS;
}
