#include "aunfold.h"

#include <iostream>

#include <set>
#include <list>

using namespace std;

  Petrigraph approximatedUnfolding(GraphGrammar& graphgrammar, GXLWriter& writer, SampleStorage& sstorage,
   double timer, bool * btimeout)
  {
    // ---------------------------------------------
    // Step (0):
    Hypergraph startgraph(graphgrammar.getStartGraph());

    // set initial depth 0
    startgraph.initDepth();
    
    // first, construct a Petrigraph with no transitions from startgraph,
    // including the identity mapping iota
    map<IDTYPE,unsigned> initialMarking;
    set<IDTYPE> startgraph_edges = startgraph.getEdges();
    set<IDTYPE> startgraph_vertices = startgraph.getVertices();

    for(set<IDTYPE>::iterator i = startgraph_edges.begin();
        i!=startgraph_edges.end();
        i++)
      initialMarking[*i] = 1;

    PTNet ptnet(startgraph_edges,list<PTNetTransition>(),initialMarking);

    map<IDTYPE,IDTYPE> identity_startgraph;
    for(set<IDTYPE>::iterator i = startgraph_edges.begin();
        i!=startgraph_edges.end();
        i++)
      identity_startgraph[*i] = *i;
    for(set<IDTYPE>::iterator i = startgraph_vertices.begin();
        i!=startgraph_vertices.end();
        i++)
      identity_startgraph[*i] = *i;

    Petrigraph P(startgraph,ptnet,graphgrammar,map<IDTYPE,map<IDTYPE,IDTYPE> >(),identity_startgraph);

    
    // --------------------------------------------
    // Step (i+1)
    map<IDTYPE,TransformationRule> rules = graphgrammar.getRules();


    bool terminated(false);
    int step = 0;

    // calculate until terminated    
    while(!terminated)
    {
      terminated = true;
      if(trace)
      {
        char graphname[256];
        sprintf(graphname,"step%i",step);
        writer.addGraph(P,graphname);
      }

      step++;
      if(!quiet)
      {
        cout << "Step " << step << ":\n" << flush;
      }
      if(debug)
      {
        cout << P << "\n";
      }

      // find a rule satisfying the conditions of def. (11)
      if(!quiet)
      {
        cout << "Searching matches" << flush;
      }

      for(map<IDTYPE,TransformationRule>::iterator rules_it = rules.begin();
          rules_it != rules.end();
          rules_it++)
      {
        if(!quiet)
        {
          cout << "." << flush;
        }
        IDTYPE r_id = (*rules_it).first; // copy ID of the rule...
        TransformationRule& r = (*rules_it).second; //.. and the rule itself
        // we call findMatching(), which will call the handler function
        // of the handler class for every matching found. This handler
        // function tests for all the conditions in (11) and stores
        // the results in its member variables. The results are
        // (a) fold or unfold or neither of these? , and
        // (b) where to fold/unfold
        FirstMatchHandler handler(P,r_id,r, sstorage);
        P.getHypergraph().findMatching(r.getLeft(),handler);
        // test the results:
        if(handler.canFold())
        {
          if(!quiet)
          {
            cout << "folding rule " << r_id << " (" << r.getName() << ")..." << flush;
          }
          P.fold(r,handler.getPhi2(),handler.getPhi());

        
          if(!quiet)
          {
            cout << "done.\n" << flush;
          }
          terminated = false;
          break; // continue with "while" loop
        }
        else if(handler.canUnfold())
        {
          if(!quiet)
          {
            cout << "unfolding rule " << r_id << " (" << r.getName() << ")..." << flush;
          }
          P.unfold(r,handler.getPhi());

          if(!quiet)
          {
            cout << "done.\n" << flush;
          }
          terminated = false;
          break; // continue with "while" loop
        }
        // else try the next rule

      }

      // step
     if(step==maxstep)
				terminated=true;

		 if(step==300)
		 {
       int i=1;
       i++; 
		 }

     // timer
		 if(timeout!=0)
		 {
			   double timeused = double(clock()) / CLOCKS_PER_SEC - timer;
			   if(timeused<0)
					 timeused=0.0;
				 if(timeused>=timeout)
				 {
					 *btimeout = true;
					 terminated = true;
				 }
		 }

				
    } // while !terminated

    // there can be different transitions with the same label and preset
    // irredundancy condition
    if(red)
			P.ReduceTransitions();

    if(!quiet)
    {
      cout << "calculation finished.\n" << flush;
    }

    return P;
  }


  // ---------------------------------------------------
  // class FirstMatchHandler:

  FirstMatchHandler::FirstMatchHandler(Petrigraph& pP,
                                       IDTYPE pr_id,
                                       TransformationRule& pr,
                                 			 SampleStorage& psstorage)
  
      :P(pP),r_id(pr_id),r(pr),
      bCanFold(false), bCanUnfold(false),
      sstorage(psstorage)
  {}

  FirstMatchHandler::~FirstMatchHandler()
  {}


  bool FirstMatchHandler::useThisMatch(std::map<IDTYPE,IDTYPE> match)
  {
    // calculate phi(E_L)
    PTMarking phi_E_L;
    set<IDTYPE> places = P.getPTNet().getPlaces();
    set<IDTYPE> E_L = r.getLeft().getEdges();
    for(set<IDTYPE>::iterator i = places.begin();
        i!=places.end();
        i++)
      phi_E_L[*i]=0;

    for(set<IDTYPE>::iterator i = E_L.begin();
        i!=E_L.end();
        i++)
      phi_E_L[match[*i]] = 1 + phi_E_L[match[*i]];

    // now the depth of vertices
		set<IDTYPE> V_L = r.getLeft().getVertices();

		bool use_depth = true;

		bool use_maxdepth = true;
		
    // does a match phi2 exist satisfying condition 6?
 		SecondMatchHandler handler(P,r_id,r,phi_E_L, match, sstorage);
		phi2 = P.getHypergraph().findMatching(r.getLeft(),handler);

    // one more time checking the depth
    // is important because of vertices
    // and of unfdepth

    for(set<IDTYPE>::iterator it = V_L.begin(); it !=V_L.end(); it++)
    {
      // depth of gvertice
	    long dep;
	    IDTYPE gvertice = match[*it];
     P.getHypergraph().getDepth(gvertice, &dep, VERTICE);

      if(!phi2.empty())
      {
				// if phi != phi2 and not depth then it doesn't pass for folding
				if((phi2[*it] != gvertice) && (dep<depth))
				{
					use_depth=false;
					break;
				}
			}

			// if more then unfdepth then it doesn't pass for unfolding
			if( (dep>=unfdepth) && (unfdepth!=-1) )
			{
				use_maxdepth=false;
				break;
			}
		}

		// and the edges
		for(set<IDTYPE>::iterator it = E_L.begin(); it !=E_L.end(); it++)
    {
      // depth of gvertice
	    long dep;
	    IDTYPE gedge = match[*it];
      P.getHypergraph().getDepth(gedge, &dep, EDGE);

      if(!phi2.empty())
      {
       // if phi != phi2 and not depth then it doesn't pass for folding
			 if((phi2[*it] != gedge) && (dep<depth))
			 {
			 	use_depth=false;
			 	break;
			 }
			}

			// if more then unfdepth then it doesn't pass for unfolding
			if( (dep>=unfdepth) && (unfdepth!=-1) )
			{
				use_maxdepth=false;
				break;
			}
		}


    if(phi2.empty() || !use_depth || nfold) // no matching found satisfying cond. 6? or depth of vertices is too small
    {
      // unfolding possible?
      // first let us check maxdepth
      if(!use_maxdepth)
				return false;
      // there is no r-labeled transition t in T_N such that preset(t) == phi_E_L ?
      list<PTNetTransition> transitions = P.getPTNet().getTransitions();
      for(list<PTNetTransition>::iterator t = transitions.begin();
          t!=transitions.end();
          t++)
      {
        // p_N_i(t) == r ?
        if(r_id == (*t).getAssociatedRule())
        {
          // preset of t == phi_E_L?
          bool equal(true);
	  //       for(map<IDTYPE,unsigned>::iterator i = phi_E_L.begin();
	  //   equal && i!=phi_E_L.end();
	  //   i++)
	  // {
          //  if((*i).second != (*t).getPreWeight((*i).first))
          //    equal = false;
          // }

          for(map<IDTYPE,IDTYPE>::iterator i = match.begin();
	      equal && i!=match.end();
              i++)
	    {
	      if(P.mu((*t).getID(),(*i).first)!=(*i).second)
                equal=false;
            }

          if(equal) return false; // found such a transition
        }
      }
      // if we reach this point, no such r-labeled transition was found.
      if(P.getPTNet().isCoverableMarking(phi_E_L))
      {
        phi = match;
        bCanUnfold = true;
        return true;
      }
      else
        return false;

    }
    else // matching found satisfying cond. 6?
    {
      if(P.getPTNet().isCoverableMarking(phi_E_L))
      {
        phi = match;
        bCanFold = true;
        return true;
      }
      else
        return false;
    }
  }


  bool FirstMatchHandler::canFold()
  {
    return bCanFold;
  }

  bool FirstMatchHandler::canUnfold()
  {
    return bCanUnfold;
  }

  map<IDTYPE,IDTYPE> FirstMatchHandler::getPhi()
  {
    return phi;
  }

  map<IDTYPE,IDTYPE> FirstMatchHandler::getPhi2()
  {
    return phi2;
  }



  // --------------------------------------------------------
  // class SecondMatchHandler
  list<CheckedFoldingsStructure> SecondMatchHandler::checkedFoldings;

  SecondMatchHandler::SecondMatchHandler(Petrigraph& pP,
                                         IDTYPE pr_id,
                                         TransformationRule& pr,
                                         PTMarking& pphi_E_L,
                                         std::map<IDTYPE,IDTYPE>& pmatchFirst,
                                         SampleStorage& psstorage)
      :P(pP),r_id(pr_id),r(pr),phi_E_L(pphi_E_L),matchFirst(pmatchFirst), sstorage(psstorage)
  {}


  bool SecondMatchHandler::useThisMatch(std::map<IDTYPE,IDTYPE> match)
  {
    // phi(E_L) is passed as a constructor parameter.
    // calculate phi2(E_L):
    PTMarking phi2_E_L;
    set<IDTYPE> places = P.getPTNet().getPlaces();
    set<IDTYPE> E_L = r.getLeft().getEdges();
    for(set<IDTYPE>::iterator i = places.begin();
        i!=places.end();
        i++)
      phi2_E_L[*i]=0;

    for(set<IDTYPE>::iterator i = E_L.begin();
        i!=E_L.end();
        i++)
      phi2_E_L[match[*i]] = 1 + phi2_E_L[match[*i]];

    // We need two _different_ matchings. phi == phi = phi2?
    bool phiequalsphi2 = true;

    for(PTMarking::iterator i = phi2_E_L.begin();
        phiequalsphi2 && i!=phi2_E_L.end();
        i++)
    {
      if(phi_E_L[(*i).first] != (*i).second)
        phiequalsphi2 = false;
    }
    if(phiequalsphi2)
      return false;

    // let us check the depth of edges
    for(set<IDTYPE>::iterator it = E_L.begin(); it !=E_L.end(); it++)
    {
      // depth of gedge
	    long dep;
	    IDTYPE gedge = match[*it];
      P.getHypergraph().getDepth(gedge, &dep, EDGE);

      // if phi != phi2 and not depth then it doesn't pass
			if((matchFirst[*it] != gedge) && (dep<depth))
				return false;
		}

		// now the depth of vertices
		set<IDTYPE> V_L = r.getLeft().getVertices();

    for(set<IDTYPE>::iterator it = V_L.begin(); it !=V_L.end(); it++)
    {
      // depth of gvertice
	    long dep;
	    IDTYPE gvertice = match[*it];
      P.getHypergraph().getDepth(gvertice, &dep, VERTICE);

      // if phi != phi2 and not depth then it doesn't pass
			if((matchFirst[*it] != gvertice) && (dep<depth))
				return false;
		}

    // Does a transition t exist such that...
    list<PTNetTransition> transitions = P.getPTNet().getTransitions();
    for(list<PTNetTransition>::iterator t = transitions.begin();
        t!=transitions.end();
        t++)
    {
      // ... p_N_i(t) == r  and ...
      if(r_id == (*t).getAssociatedRule())
      {
        // ...preset of t == phi2_E_L and ...
        bool equal(true);
        //for(map<IDTYPE,unsigned>::iterator i = phi2_E_L.begin();
        //    equal && i!=phi2_E_L.end();
        //    i++)
	// {
        //  if((*i).second != (*t).getPreWeight((*i).first))
        //    equal = false;
        // }

        for(map<IDTYPE,IDTYPE>::iterator i = match.begin();
              equal && i!=match.end();
              i++)
	   {
	     if(P.mu((*t).getID(),(*i).first)!=(*i).second)
               equal=false;
           }

        if(equal)
        {
          // ... for all e in phi_E_L: e in preset(t) or t <N e
          bool e_condition_satisfied = true;
          for(PTMarking::iterator e = phi_E_L.begin();
              e_condition_satisfied && e!=phi_E_L.end();
              e++)
          {
            if(      0 <(*e).second
                     && 0==(*t).getPreWeight((*e).first) // not in preset
                     && !P.getPTNet().inCausalityRelation((*t).getID(),(*e).first)) // not t <N e
              e_condition_satisfied = false;
          }
          // found such a transition
          if(e_condition_satisfied)
          {
						// make a copy of petrigraph
						Petrigraph pg_copy(P);
						if(checked(r, match, matchFirst))
							return false;

            pg_copy.unrealFolding();
						pg_copy.fold(r ,match, matchFirst);

						if(sstorage.checkPetrigraph(pg_copy))
							return true;
						else
						{
							CheckedFoldingsStructure cfs(r, match, matchFirst);
							checkedFoldings.push_back(cfs);

						return false;
						}
          }
        }
      }
    }
    return false;

  }

/** check if this folding variant was already checked */
bool SecondMatchHandler::checked(TransformationRule & r, map<IDTYPE, IDTYPE> & match, map<IDTYPE, IDTYPE> & matchFirst)
{
	 for(list<CheckedFoldingsStructure>::iterator it = checkedFoldings.begin();
									 it != checkedFoldings.end(); it++)
	 {
		  if(r.getID()!=it->rid)
				continue;

			if(!(match==it->match))  
				continue;

			if(!(matchFirst==it->matchFirst))
				continue;

			return true;	
	 }

	 return false;
}
