#ifndef AUNFOLD_H
#define AUNFOLD_H

#include "Petrigraph.h"
#include "GraphGrammar.h"
#include "Hypergraph.h"
#include "Globals.h"
#include "GXLWriter.h"
#include "SampleStorage.h"
#include <map>


  /** given a graph grammar, aunfold calculates its appoximated unfolding. */;
   Petrigraph approximatedUnfolding(GraphGrammar& graphgrammar, GXLWriter& writer, SampleStorage& sstorage,
						 double timer, bool * btimeout);


   // handles the "phi" matches:
   class FirstMatchHandler : public MatchHandler
     {
     public:
       FirstMatchHandler(Petrigraph& pP,
			 IDTYPE pr_id,
			 TransformationRule& pr,
 			 SampleStorage& psstorage);

			 virtual ~FirstMatchHandler();
       bool useThisMatch(std::map<IDTYPE,IDTYPE> match);
       bool canFold();
       bool canUnfold();
       /** PRE: (canFold() || canUnfold()) */
       std::map<IDTYPE,IDTYPE> getPhi();
       /** PRE: canFold() */
       std::map<IDTYPE,IDTYPE> getPhi2(); 

     private:
       Petrigraph& P;
       IDTYPE r_id;
       TransformationRule& r;
       
       bool bCanFold;
       bool bCanUnfold;
       
       std::map<IDTYPE,IDTYPE> phi;
       std::map<IDTYPE,IDTYPE> phi2;

 			 SampleStorage& sstorage;
     }
   ;
   

   struct CheckedFoldingsStructure
   {
     CheckedFoldingsStructure(TransformationRule & r,
															 map<IDTYPE, IDTYPE> & match,
															 map<IDTYPE, IDTYPE> & matchFirst)
		 {
			 this->rid = r.getID();
			 this->match = match;
			 this->matchFirst = matchFirst;
		 }
		 
		 IDTYPE rid;
		 map<IDTYPE, IDTYPE> match;
		 map<IDTYPE, IDTYPE> matchFirst;
   };

   // handles the "phi2" (or " phi' ") matches. Used only by FirstMatchHandler.
   class SecondMatchHandler : public MatchHandler
     {
     public:
       SecondMatchHandler(Petrigraph& pP,
			  IDTYPE pr_id,
			  TransformationRule& pr,
			  PTMarking& pphi_E_L,
			  std::map<IDTYPE,IDTYPE>& pmatchFirst,
			  SampleStorage& psstorage);
       // tests for condition (6)
       bool useThisMatch(std::map<IDTYPE,IDTYPE> match);

       bool checked(TransformationRule & r, std::map<IDTYPE, IDTYPE> & match, std::map<IDTYPE, IDTYPE> & matchFirst);

     private:
       Petrigraph& P;
       IDTYPE r_id;
       TransformationRule& r;
       PTMarking& phi_E_L;
  		 std::map<IDTYPE,IDTYPE> matchFirst;
			 SampleStorage& sstorage;
       
     public: 
       static std::list<CheckedFoldingsStructure> checkedFoldings;
     };
   
 
   


#endif
